# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Wed Oct  2 19:57:54 2019

Column base joint according to EN 1993-1-8

@author: kmela
"""

import math

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as lines


    
from metku.eurocodes.en1993 import constants
from metku.eurocodes.en1993 import en1993_1_1
from metku.materials import steel_data
import metku.eurocodes.en1992.en1992_1_1 as en1992
import metku.eurocodes.en1992.constants as en1992_const
from metku.sections.steel.ISection import HEA
from metku.structures.steel.plates import RectPlate
from metku.materials.steel_data import Steel
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import TENSION_ROW, COMPRESSION_ROW, SHEAR_ROW, COMBI_ROW
from metku.eurocodes.en1993.en1993_1_8.en1993_1_8 import TStubEndPlate, BoltRow

ROW_OUTSIDE_COLUMN_FLANGE = "Row outside column flange"
ROW_INSIDE_COLUMN_FLANGE = "Row inside column flange"


# Anchor bolts
    
HPM_L = {16:{"d":16,"As":157,"L":280,"washer":[40,6],"dh":38,"k":10, "material": "B500B"},
             20:{"d":20,"As":245,"L":350,"washer":[44,6],"dh":46,"k":12, "material": "B500B"},
             24:{"d":24,"As":352,"L":430,"washer":[56,6],"dh":55,"k":13,"material": "B500B"},
             30:{"d":30,"As":561,"L":500,"washer":[65,8],"dh":70,"k":15, "material": "B500B"},
             39:{"d":39,"As":976,"L":700,"washer":[90,10],"dh":90,"k":18, "material": "B500B"},
             }

class AnchorBolt:
    """ For anchor bolt data """
    
    def __init__(self,bolt_type=HPM_L[24]):
        """ Constructor
            input:
                bolt_type .. from a list of bolts above
                
            attributes:
                d .. diameter of the shank (mm)
                As .. stress area of the threaded shank (mm2)
                L .. length of the bolt (mm)
                washer .. [0] washer diameter (mm)
                          [1] washer thickness (mm)
                dh .. diameter of the head [mm]
                k .. thickness of the head (mm)
                material .. 
        """
        
        self.d = bolt_type["d"]
        self.As = bolt_type["As"]
        self.L = bolt_type["L"]
        self.washer = bolt_type["washer"]
        self.dh = bolt_type["dh"]
        self.k = bolt_type["k"]
        self.material = Steel(bolt_type["material"])
        
class AnchorBoltRow:
    """ Class for anchor bolt rows 
        
        Methods and functionality:
            - compute tension resistance of each component
            - compute stiffness of each component
        
        The components are:
            - end plate in bending
            - column flange in bending
            - column web in tension
            - beam web in tension
    
    """
    
    def __init__(self,bolt,p,z,loc=ROW_OUTSIDE_COLUMN_FLANGE,joint=None, row_type=TENSION_ROW):
        """ Constructor
            input:
                bolt .. Bolt class object defining the bolt
                p .. array of distances between adjacent bolt centroids
                z .. horizontal position of bolt row (origin is at the centroid of the column section)
                loc .. position of the row with respect to column flange                
                joint .. the joint object to which the bolt row belongs
                row_type .. TENSION_ROW (default)
        """
        
        self.type = row_type
        self.bolt = bolt
        #self.p = p
        self.z = z
        self.loc = loc        
        
        """ Location of bolt row in a group. This varies,
            if the row belongs to several groups        
        """
        self.group_loc_flange = None
        self.group_loc_plate = None
        
        self.joint = joint
        
        # Edge distance
        self._e1 = 0.5*self.joint.column.h+self.joint.etop - z
        #self.y = y
        """ p can be a single number, in which case there are two bolts
            in the row
            
            if p is an array, the number of bolts is length of p + 1
        """
        if isinstance(p,float) or isinstance(p,int):
            self.bolts = 2
        else:
            self.bolts = len(p)+1
        
        """ Initialize effective lengths corresponding to
            column flange in bending (leff_flange) and
            end plate in bending (leff_plate), when
            the row is treated as an individual row
            
            first element is for Mode 1 and second for Mode 2
        """
        #self.Tstub_col_flange = TStubColumnFlange([bolt,bolt],flange_loc,self.joint.col.material,tf=10,e=20,emin=20,m=10)
        self.Tstub_col_flange = TStubColumnFlange(self,self.joint.col.material,tf=10,e=20,emin=20,m=10)
        self.Tstub_end_plate = TStubEndPlate(self,self.joint.end_plate.material,tf=10,emin=20,m=10,e=20,w=20,bp=self.joint.bp)
        #self.leff_flange = [0.0,0.0]
        #self.leff_plate = [0.0,0.0]
        
        
        
        self.FtRd = -1.0
    
        # Dictionary of stiffness factors
        self.stiffness_factors = {"col_web": np.inf,
                                  "col_flange": np.inf,
                                  "plate":np.inf,
                                  "bolt":1.6*self.bolt.As/self.Lb}
        
        # List of row groups that contain this row
        # this list appended when bolt row groups are created.
        self.groups = []
    
        # INITIALIZATION: CALCULATE VARIOUS GEOMETRIC QUANTITIES
        
        """ Effective length of column flange """
        #self._leff_nc_flange = self.Tstub_col_flange.leff_nc()
        #self._leff_cp_flange = self.Tstub_col_flange.leff_cp()
        
        """ Effective length of end plate """
        #self._leff_nc_flange = self.Tstub_end_plate.leff_nc()
        #self._leff_cp_flange = self.Tstub_end_plate.leff_cp()
    
    def __repr__(self):
        
        return "Bolt row"
    
    
    def info(self):
        """ Prints info """
        
        print("Bolt row:")
        print("Location (flange): {0:s}".format(self.loc_col_flange))
        print("Location (end plate): {0:s}".format(self.loc_end_plate))
        
    @property
    def bolt_size(self):
        return self.bolt.d
    
    @bolt_size.setter
    def bolt_size(self,val):
        self.bolt.size = val
        
    @property
    def z_top(self):
        """ Distance of the bolt row from the top of the end plate """
        return self._z_top
    
    @z_top.setter
    def z_top(self,val):
        """ Distance of the bolt row from the top of the end plate """
        self._z_top = val
        self.z = 0.5*self.joint.beam.h+self.joint.etop - val
        
    @property
    def Lb(self):
        """ Ruuvirivin venymäpituus """
        return self.joint.tp + self.joint.col.tf + 2*self.bolt.washer_t + 0.5*(self.bolt.head_t + self.bolt.nut_t)
    
    @property
    def h(self):
       """ Distance from the center of compression of the joint """
       return self.z-self.joint.zc 
    
    @property
    def emin_flange(self):
        """ Distance 'emin' used for T-stubs for the component
            'column flange in bending'
        """
        """ If the plate is narrower than column flange,
            take 'emin' to be the distance to plate edge.
                
            Otherwise, use the distance to column flange edge.
        """
        if self.joint.bp < self.joint.col.b:
            emin = self.joint.ebolts
        else:
            emin = self.joint.ebolts - 0.5*(self.joint.bp-self.joint.col.b)
        
        return emin
    
    @property
    def emin_plate(self):
        """ Distance 'emin' used for T-stubs for the component
            'end plate in bending'
        """
        if self.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
            """ For the row in the plate extension, use the
                distance to plate end, ex.
            """
            emin = self.ex
        else:
            """ If the plate is narrower than column flange,
                take 'emin' to be the distance to plate edge.
                
                Otheriwse, use the distance to column flange edge.
            """
            if self.joint.bp < self.joint.col.b:
                emin = self.joint.ebolts
            else:
                emin = self.joint.ebolts - 0.5*(self.joint.bp-self.joint.col.b)
        
        return emin
    
    @property
    def ex(self):
        """ Distance from bolt row center to top edge of the end plate
        """
        
        return 0.5*self.joint.beam.h + self.joint.etop - self.z
    
    @property
    def mx(self):
        """ Distance from bolt row center to top of the top flange
            of the beam. This is used for bolt rows outside the
            beam tension flange.
        """
        
        return self.z - 0.5*self.joint.beam.h - 0.8*math.sqrt(2)*self.joint.weld_f
    
    @property
    def w(self):
        """ Horizontal distance between bolt holes """
        return self.joint.p        
        
    @property
    def col_flange_m(self):
        """ Distance from column web to bolt hole center """
        
        return 0.5*(self.w-self.joint.col.tw)-0.8*self.joint.col.r
    
    @property
    def col_flange_e(self):
        """ Distance of column flange edge from the bolt hole center """
        if self.joint.bp < self.joint.col.b:
            e = self.joint.ebolts + 0.5*(self.joint.col.b-self.joint.bp)
        else:
            e = self.joint.ebolts - 0.5*(self.joint.bp-self.joint.col.b)
        
        return e
    
    @property
    def end_plate_m(self):
        """ Distance from beam web to bolt hole center """
        if self.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
            m = self.mx
        else:
            m = 0.5*(self.w-self.joint.beam.tw)-0.8*math.sqrt(2)*self.joint.weld_w 
        return m
    
    @property
    def end_plate_m2(self):
        """ Distance from beam flange to bolt hole center """
        
        return 0.5*self.joint.beam.h-self.joint.beam.tf-0.8*math.sqrt(2)*self.joint.weld_f-self.z
    
    @property
    def end_plate_e(self):
        """ Distance of end plate edge from the bolt hole center """
        
        return self.joint.ebolts
    
    @property
    def top_flange_stiffener_m2(self):
        """ Distance of bolt row from top flange stiffener """
        zbeam = 0.5*self.joint.beam.h
        if self.z > zbeam:
            """ Row is above the stiffener """
            zweld = zbeam - 0.5 * self.joint.beam.tf + \
                    + 0.5 * self.joint.stiffeners['col_top_flange'].khp + \
                    + 0.8 * math.sqrt(2) * self.joint.stiffeners['col_top_flange'].weld_size
            return self.z-zweld
        else:
            """ Row is below the stiffener """
            zweld = zbeam - 0.5 * self.joint.beam.tf + \
                    - 0.5 * self.joint.stiffeners['col_top_flange'].khp + \
                    - 0.8 * math.sqrt(2) * self.joint.stiffeners['col_top_flange'].weld_size
            
            return zweld-self.z
    
    def leff_flange(self):
        """ Effective length of column flange in bending component """
        return self.Tstub_col_flange.leff()
    
    def leff_plate(self):
        """ Effective length of end plate in bending component """
        return self.Tstub_end_plate.leff()
    
    def leff_flange_group(self):
        """ Effective length of column flange in bending for the row
            when considered to be in a group
            
            IDEA:
                for all groups that have the row:
                    1) get row location in that group
                    2) calculate effective width for the row
                        considered as in a group and with the
                        obtained location.
                    3) add the calculated effective width to the
                        list of effective widths
                    4) return minimum
    
            QUESTION: how to know, which groups have the row
                in them.
        """
        leff = 1e5
        """ The for loop is over all the groups that involve this bolt row
            (the groups are stored in 'groups' attribute)
        """
        for group in self.groups:
            """ The '_leff_rows' attribute of the column flange T-stub
                for the group is a dict with BoltRow objects as keys. Using
                'self' will get the effective lengths of this bolt row
                appearing in the given group. Then, take the minimum of those values.
            """
            #leff_g = min(group.Tstub_col_flange._leff_rows[self].values())
            leff_g = min(group.Tstub_col_flange.leff_cp_row(self),group.Tstub_col_flange.leff_nc_row(self))
            if leff_g > 0.0:
                """ If the group does not have column flange in bending
                    component, then leff_g = 0 and that group can be skipped.
                    Otherwise, update the effective length.
                """
                leff = min(leff,leff_g)        
            
        return leff
        
    
    def leff_plate_group(self):
        """ Effective length of end plate in bending component
            for the row when considered to be in a group
        """
        leff = 1e5
        for group in self.groups:
            #leff_g = min(group.Tstub_end_plate._leff_rows[self].values())
            
            # If the group does not have the end plate in bending component,
            # return a large value.
            
            if not group.Tstub_end_plate is None:
                leff_g = min(group.Tstub_end_plate.leff_cp_row(self),group.Tstub_end_plate.leff_nc_row(self))                
            else:
                leff_g = leff
            if leff_g > 0.0:
                leff = min(leff,leff_g)
            
        return leff
    
    def k3(self,verb=False):
        """ Stiffness factor for column web in tension component """
        
        #beff_row = self.leff_flange()
        # NOTE! Here, the _leff attribute of the T-stub corresponds to the 
        # effective lengths calculated during bending resistance evaluation.
        # This can be a problem if the bending resistance has not been
        # calculated before rotational stiffness, or when calculating derivatives
        # in optimization using finite differences.
        #
        # beff_row = min(self.Tstub_col_flange._leff)
        
        #self.Tstub_col_flange.tf = self.joint.col.tf
        #self.Tstub_col_flange.emin = self.emin_flange
        #self.Tstub_col_flange.m = self.col_flange_m
        #self.Tstub_col_flange.e = self.col_flange_e
        
        beff_row = min(self.Tstub_col_flange.leff_cp,self.Tstub_col_flange.leff_nc)
        
        #beff_group = 1e5
        beff_group = self.leff_flange_group()
        
        beff = min(beff_row,beff_group)
        
        if verb:
            print(f'beff_row = {beff_row:4.2f} mm')
            print(f'beff_group = {beff_group:4.2f} mm')
        
        return cm.column_web_tension_k(self.joint.col,beff,verb)
    
    
    def k4(self,verb=False):
        """ Stiffness factor for column flange in bending component """
        
        """ This way picks that effective length that corresponds to the
            governing failure mode.
        beff_row = self.leff_flange()
        """
        
        """ This way takes the smallest effective length for the row """
        #beff_row = min(self.Tstub_col_flange._leff)
        #self.Tstub_col_flange.tf = self.joint.col.tf
        #self.Tstub_col_flange.emin = self.emin_flange
        #self.Tstub_col_flange.m = self.col_flange_m
        #self.Tstub_col_flange.e = self.col_flange_e
        
        beff_row = min(self.Tstub_col_flange.leff_cp,self.Tstub_col_flange.leff_nc)
        #beff_group = 1e5
        beff_group = self.leff_flange_group()
        leff = min(beff_row,beff_group)        
        
        
        m = self.Tstub_col_flange.m
        return cm.column_flange_bending_k(self.joint.col,leff,m,verb)
        
    def k5(self,verb=False):
        """ Stiffness factor for end plate in bending component """
        
        #beff_row = self.leff_plate()
        #beff_row = min(self.Tstub_end_plate._leff)
        #self.Tstub_end_plate.m2 = self.end_plate_m2
        
        #self.Tstub_end_plate.bp = self.joint.bp
        #self.Tstub_end_plate.tf = self.joint.tp
        #self.Tstub_end_plate.w = self.p
        
        #if self.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
        #    self.Tstub_end_plate.emin = self.ex
        #    self.Tstub_end_plate.e = self.end_plate_e
        #    self.Tstub_end_plate.m = self.mx
        #else:
        #    self.Tstub_end_plate.emin = self.emin_plate
        #    self.Tstub_end_plate.e = self.end_plate_e
        #    self.Tstub_end_plate.m = self.end_plate_m
        #    self.Tstub_end_plate.m2 = self.end_plate_m2
        
        beff_row = min(self.Tstub_end_plate.leff_cp,self.Tstub_end_plate.leff_nc)
        
        
        #beff_group = 1e5
        beff_group = self.leff_plate_group()
        
        #print(f'beff (row) = {beff_row:4.2f}')
        #print(f'beff (group) = {beff_group:4.2f}')
        
        leff = min(beff_row,beff_group)                
        
        m = self.Tstub_end_plate.m
        
        #print(beff_row,beff_group,m)
        return cm.end_plate_bending_k(self.joint.tp,leff,m,verb)
    
    def k10(self,verb=False):
        """ Stiffness factor for bolts in tension """
        
        k10 = 1.6*self.bolt.As/self.Lb
        
        return k10
        
    
    def column_web_in_tension(self,verb=False):
        """ Resistance of component """
        #self.Tstub_col_flange.tf = self.joint.col.tf
        #self.Tstub_col_flange.emin = self.emin_flange
        #self.Tstub_col_flange.m = self.col_flange_m
        #self.Tstub_col_flange.e = self.col_flange_e
        
        """ The effective length of the component
            corresponds to the effective length of the column flange.
            This is obtained only after the resistance of the column
            flange has been determined, because the governing failure
            mode also defines the effective width
        """
        #l_eff = self.Tstub_col_flange.leff(self.loc_col_flange)
        
        #l_eff = [0,0]
        #l_eff[0] = self.Tstub_col_flange.leff_1()
        #l_eff[1] = self.Tstub_col_flange.leff_2()
        
        l_eff = self.Tstub_col_flange.leff()
        #print(self.Tstub_col_flange.leff)
        
        F_t_wc_Rd = cm.column_web_tension(self.joint.col, l_eff, self.joint.beta,verb)
        
        if verb:
            print("COlUMN WEB IN TENSION:")
            print("l_eff  = {0:4.2f} mm".format(l_eff))
            #print("l_eff (Mode 2) = {0:4.2f} mm".format(self.Tstub_col_flange._leff[1]))
            print("Ft_wc_Rd = {0:4.2f} kN".format(F_t_wc_Rd*1e-3))
        
        return F_t_wc_Rd
    
    def column_flange_in_bending(self,verb=False):
        """ Resistance of component """
        
        """ Set properties for the T-stub
            This is basically for optimization purposes,
            because the dimensions will change
        """
        #self.Tstub_col_flange.tf = self.joint.col.tf
        #self.Tstub_col_flange.emin = self.emin_flange
        #self.Tstub_col_flange.m = self.col_flange_m
        #self.Tstub_col_flange.e = self.col_flange_e
        if self.joint.stiffeners['col_top_flange'] is not None:
            self.Tstub_col_flange.m2 = self.top_flange_stiffener_m2
        """ Determine effective length 
            The 'leff_1' and 'leff_2' methods store the
            effective length of each failure mode to the T-stub,
            so it can be
        """
        
        """ Calculate the effective length of the T-stub
            The effective length will be stored in the T-stub object.
        """
        l_eff = [0.0,0.0]
        l_eff[0] = self.Tstub_col_flange.leff_1
        l_eff[1] = self.Tstub_col_flange.leff_2
        
        if verb:
            print("COLUMN FLANGE IN BENDING:")
            print("l_eff (Mode 1) = {0:4.2f} mm".format(l_eff[0]))
            print("l_eff (Mode 2) = {0:4.2f} mm".format(l_eff[1]))
        
        """ The resistance is based on the T-stub """
        F_t_cf_Rd = cm.column_flange_bending(self.Tstub_col_flange,verb)
        
        if verb:            
            print("Ft_cf_Rd = {0:4.2f} kN".format(F_t_cf_Rd*1e-3))
        
        return F_t_cf_Rd
    
    def end_plate_in_bending(self,verb=False):
        """ Resistance of component """
        
        """
        self.Tstub_end_plate.bp = self.joint.bp
        self.Tstub_end_plate.tf = self.joint.tp
        self.Tstub_end_plate.w = self.p
        
        if self.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
            self.Tstub_end_plate.emin = self.ex
            self.Tstub_end_plate.e = self.end_plate_e
            self.Tstub_end_plate.m = self.mx
        else:
            self.Tstub_end_plate.emin = self.emin_plate
            self.Tstub_end_plate.e = self.end_plate_e
            self.Tstub_end_plate.m = self.end_plate_m
            self.Tstub_end_plate.m2 = self.end_plate_m2
        """
        
        l_eff = [0.0,0.0]
        l_eff[0] = self.Tstub_end_plate.leff_1
        l_eff[1] = self.Tstub_end_plate.leff_2
        
        if verb:
            print("END PLATE IN BENDING:")
            self.Tstub_end_plate.info()
            print("l_eff (Mode 1) = {0:4.2f} mm".format(l_eff[0]))
            print("l_eff (Mode 2) = {0:4.2f} mm".format(l_eff[1]))
        
        F_t_ep_Rd = cm.end_plate_bending(self.Tstub_end_plate,verb)
        
        return F_t_ep_Rd
    
    def beam_web_in_tension(self,verb=False):
        """ Resistance of component """
        """
        self.Tstub_end_plate.tf = self.joint.tp
        self.Tstub_end_plate.w = self.p
        
        if self.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
            self.Tstub_end_plate.emin = self.ex
            self.Tstub_end_plate.e = self.ex
            self.Tstub_end_plate.m = self.mx
        else:
            self.Tstub_end_plate.emin = self.emin_plate
            self.Tstub_end_plate.e = self.end_plate_e
            self.Tstub_end_plate.m = self.end_plate_m
            self.Tstub_end_plate.m2 = self.end_plate_m2
        """
        
        if self.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
            F_t_bw_Rd = 1e8
        else:        
            l_eff = self.Tstub_end_plate.leff()
            #l_eff[0] = self.Tstub_end_plate.leff_1(self.loc_end_plate)
            #l_eff[1] = self.Tstub_end_plate.leff_2(self.loc_end_plate)
        
            F_t_bw_Rd = cm.beam_web_tension(self.joint.beam, l_eff)
        
        if verb:
            print("BEAM WEB IN TENSION:")
            if self.loc_end_plate == ROW_OUTSIDE_BEAM_TENSION_FLANGE:
                print("Row outside beam tension flange: no beam web in tension.")
            else:
                print("l_eff = {0:4.2f} mm".format(l_eff))
                print("Ft_bf_Rd = {0:4.2f} kN".format(F_t_bw_Rd*1e-3))
        
        return F_t_bw_Rd
    
    def tension_resistance(self,verb=False):
        """ Tension resistance of individual row, not considered
            as a group.
            
            Evaluate first the components
            'column flange in bending' and
            'end plate in bending', because
            they define the effective lengths also used in
            'column web in tension' and
            'beam web in tension' components
        """
        
        Ft_i_Rd = np.array([0.0,0.0,0.0,0.0])
        components = ['Column flange in bending',
                      'End plate in bending',
                      'Column web in tension',
                      'Beam web in tension'
                      ]
        
        Ft_i_Rd[0] = self.column_flange_in_bending(verb)
        Ft_i_Rd[1] = self.end_plate_in_bending(verb)
        Ft_i_Rd[2] = self.column_web_in_tension(verb)
        Ft_i_Rd[3] = self.beam_web_in_tension(verb)
            
        min_comp = np.argmin(Ft_i_Rd)
        
        self.FtRd = Ft_i_Rd[min_comp]
        
        if verb:
            print("Ft_Rd = {0:4.2f} kN ({1:s})".format(self.FtRd*1e-3,components[min_comp]))
        
        return self.FtRd
    
    def keff(self):
        """ Effective stiffness of bolt row 
            EN 1993-1-8, Eq. (6.30)
            
            Requires that the stiffness factors have been evaluated
            and stored in 'stffiness_factors' dict.
        """
        
        # Inverse of keff
        keff_inv = 0.0
        
        self.stiffness_factors['col_web'] = self.k3()
        self.stiffness_factors['col_flange'] = self.k4()
        self.stiffness_factors['plate'] = self.k5()
        self.stiffness_factors['bolt'] = self.k10()
        
        #print(self.stiffness_factors)
        
        for k in self.stiffness_factors.values():
            keff_inv += 1/k
        
        return 1/keff_inv

class ColumnBase:
    
    
    def __init__(self,column_profile, plate_edges, plate_thickness, anchor_bolts, bolt_z=None, bolt_w=100, concrete="C25/30"):
        """ Constructor
            Input:
                column_profile .. SteelSection class object
                plate_edges .. list of edge distances
                               [0] .. distance from left column flange to edge
                               [1] .. distance from right column flange to edge
                               [2] .. distance from top edge to column flange
                               [3] .. distance from bottom edge to column flange
                plate_thickness .. thickness of the base plate [mm]
                anchor_bolts
                        
        """

        self.column = column_profile        
        self._c = plate_edges
        h = column_profile.h
        b = column_profile.b
        hp = h + self._c[0] + self._c[1]
        bp = b + self._c[2] + self._c[3]
        base_plate = RectPlate(bp,hp,plate_thickness,material="S355")
        self.plate = base_plate
        self.anchor_bolt = anchor_bolts
        self.w = bolt_w
        
        # By default, place anchor bolt rows halfway at outstand part of the base plate
        if bolt_z is None:
            bolt_z = [-0.5*h-0.5*self._c[0], 0.5*h+0.5*self._c[1]]
        
        self.zb = bolt_z
        self.concrete = en1992_const.Concrete(concrete)        
        self.Ac0 = 1.0
        self.Ac1 = 1.0
        self.flange_weld = 5.0
        self.web_weld = 5.0
        self.bolt_rows = []
        self.add_bolt_row(anchor_bolts, bolt_z[0], self.w)
        self.add_bolt_row(anchor_bolts, bolt_z[1], self.w)
        self.beta_j = 2.0/3.0
        
    
    def add_bolt_row(self,bolt,z,p):
        """ Input:
            bolt .. bolt type
            z .. position of bolts relative to centroid of column
                 z < 0: left side
                 z > 0: right side
            p .. distance between bolts
        """
        #(self,bolt,p,z,flange_loc=INNER_ROW, plate_loc=OTHER_INNER_ROW,joint=None, row_type=TENSION_ROW)
        
        if z < -0.5*self.column.h:
            loc = ROW_OUTSIDE_COLUMN_FLANGE
            row_type = TENSION_ROW
        elif z < 0:
            loc = ROW_INSIDE_COLUMN_FLANGE
            row_type = TENSION_ROW
        else:
            loc = ROW_OUTSIDE_COLUMN_FLANGE
            row_type = COMPRESSION_ROW
        
        new_row = AnchorBoltRow(bolt,p,z,loc=loc,joint=self,row_type=row_type)
        
        self.bolt_rows.add(new_row)
    
    @property
    def fcd(self):
        """ Design compression strength of concrete """        
        return self.concrete.fcd()
        #return en1992.fcd(en1992.concrete[self.concrete]["])

    @property
    def fjd(self):
        """ Design compression strength of concrete """                     
        # Strength of the foundation
        return self.beta_j*self.fcd        

    @property
    def tp(self):
        """ Thickness of end plate """
        return self.plate.t
        #return en1992.fcd(en1992.concrete[self.concrete]["])

    @property
    def fyp(self):
        """ Yield strength of end plate """
        return self.plate.material.fy
        #return en1992.fcd(en1992.concrete[self.concrete]["])
        
    @property
    def c(self):
        """ Length 'c' of EN 1993-1-8, Eq. (6.5) """
        return self.tp*math.sqrt(self.fyp/3/self.fjd)
    
    
    def beff(self,c):
        """ Effective width """
        return 2*c+self.column.tf
    
    
    def leff(self,c):
        """ Effective length """
        return 2*c+self.column.b
    
    def bolt_coord(self):
        """ Returns list of tuples that are the xy-coordinates of the bolts """
        coords = []
        
        ypos = 0.5*self.p
        yneg = -ypos
        for xb in self.xb:    
            coords.append((xb,yneg))
            coords.append((xb,ypos))
            
        return coords
    
    def Fc_pl_Rd(self,verb=False):
        """ Compression strength of T-stub 
            EN 1993-1-8, Eq. (6.4)
        """
        c = self.c
        beff = self.beff(c)
        leff = self.leff(c)
        fjd = self.fjd
        
        FRd = fjd*beff*leff
        
        if verb:
            print("Compression resistance of T-stub:")
            print(f"c = {c:4.2f} mm")
            print(f"beff = {beff:4.2f} mm")
            print(f"fjd = {fjd:4.2f} MPa")
            print(f"FC,Rd = {FRd*1e-3:4.2f} kN")
        
        return FRd

    def Fc_fc_Rd(self,verb=False):
        """ Column flange and web in compression 
            EN 1993-1-8 6.2.6.7
        """
        McRd = self.column.MRd
        Fc = McRd/(self.column.h-self.column.tf)
        
        if verb:
            print("Column flange and web in compression")
            print(f"McRd = {McRd*1e-6:4.2f} kNm")
            print(f" h = {self.column.h:4.2f} mm")
            print(f" tf = {self.column.tf:4.2f} mm")
            print(f" Fc_fc_Rd = {Fc*1e-3:4.2f} kN")
        
        return Fc
    
    def FCRd(self,verb=False):
        """ Compression strength of the connection """
        
        FcplRd = self.Fc_pl_Rd(verb)
        FcfcRd = self.Fc_fc_Rd(verb)
        
        FC = min(FcplRd,FcfcRd)
        
        return FC
    
    def Ft_ep_Rd(self,verb=False):
        """ Tension side: bending of base plate.
            EN 1993-1-8, 6.2.6.5. Base plate is treated as end plate.
        """
        
        ex = 0.5*self.plate.h+self.xb[0]
        mx = 0.5*self.column.h+0.8*self.flange_weld*math.sqrt(2) + self.xb[0]
        
        T = TStubEndPlate(self.bolt_rows[0],
                          material=self.plate.material,
                          tf=self.plate.t,
                          emin=ex,
                          m=mx,
                          e=ex,
                          w=self.p,
                          bp=self.plate.b)
        
        F12Rd = T.F_T_12_Rd()
        F3Rd = T.F_T_3_Rd()
        
        
    
    def draw(self, name=""):
        """ Draw the connection """
        fig, ax = plt.subplots(1,2)
        ymax = 150
        
        self.column.draw(axes=ax[1])
        
        # Side view:
        # Origin is located at the bottom of the steel profile center line
        hp, bp, tp = self.plate.h, self.plate.b, self.plate.t
        base_plate = patches.Rectangle((-0.5*hp, -tp), width=hp, height=tp, fill=False, hatch='\\')
        
        base_plate_plan = patches.Rectangle((-0.5*hp, -0.5*bp), width=hp,
                                       height=bp, fill=False, hatch='//')
        
        
        # Draw bolts
        dbolt = self.anchor_bolt.d
        lbolt = self.anchor_bolt.L
        left_bolt = patches.Rectangle((self.xb[0]-0.5*dbolt, -tp-lbolt), width=dbolt,
                                       height=lbolt+tp, fill=True)
        
        right_bolt = patches.Rectangle((self.xb[1]-0.5*dbolt, -tp-lbolt), width=dbolt,
                                       height=lbolt+tp, fill=True)
        """
        bot_flange = patches.Rectangle((-0.5*self.bb, 0), width=self.bb,
                                       height=self.tb, fill=False, hatch='\\')
        web = patches.Rectangle((-0.5*self.tw, self.tb), width=self.tw,
                                height=self.hw, fill=False, hatch='\\')
    
        top_flange = patches.Rectangle((-0.5*self.bt, self.h-self.tt),
                                       width=self.bt, height=self.tt,
                                       fill=False, hatch='\\')
        """
        ax[0].add_patch(base_plate)
        h = self.column.h
        tf = self.column.tf
        ax[0].vlines([-0.5*h,-0.5*h+tf,0.5*h,0.5*h-tf],0,ymax)
        ax[0].vlines(0,0,ymax,linestyles='dashdot')
        ax[0].add_patch(left_bolt)
        ax[0].add_patch(right_bolt)
        ax[0].set_xlim(-0.5*hp, 0.5*hp)
        if self.anchor_bolt is None:
            ymin = -self.plate.t
        else:
            ymin = -self.anchor_bolt.L
        
        ax[0].set_ylim(ymin, ymax)
    
        
        ax[0].set_aspect('equal')
        
        xtick = ax[0].get_xticks()
        #print(xtick)
        
        """ Draw connection from above """
        ax[1].add_patch(base_plate_plan)
        
        # Draw column profile
        
        
        # Draw bolts
        xy_bolts = self.bolt_coord()

        for xy in xy_bolts:
            ax[1].add_patch(patches.Circle(xy,radius=0.5*dbolt,linestyle='solid'))

        
        ax[1].set_xticks(xtick)
        #ax[1].set_xlim(-0.5*hp, 0.5*hp)
        ax[1].set_ylim(-0.5*hp, 0.5*hp)
        ax[1].set_aspect('equal')
        
        return ax
        
if __name__ == '__main__':
    
    col = HEA(240)
    #w = col.b + 20
    #d = col.h + 140
    #plate = RectPlate(width=d,depth=d,thickness=30,material="S355")
        
    
    #conc = en1992.concrete["C25/30"]
    
    col.Ned = -263.22e3
    col.Med[0] = 39.57e6
    col.Ved[1] = 16.70e3
    
    c = [100,100,50,50]
    tp = 40
    bolts = AnchorBolt(HPM_L[24])

    joint = ColumnBase(col,plate_edges = c,\
                       plate_thickness = tp,\
                       anchor_bolts = bolts,\
                       bolt_p = 100,\
                       concrete="C30/37")
        
    ax = joint.draw()
    
    #concrete = en1992_const.Concrete("C25/30")
    #fcd = concrete.fcd
    #beta_j = 2.0/3.0
    # Strength of the foundation
    #fjd = beta_j*fcd
    
    """
    print("fcd = {0:4.2f} MPa".format(joint.fcd))
    print("fjd = {0:4.2f} MPa".format(joint.fjd))
    
    # Weld sizes
    joint.flange_weld = 6
    joint.web_weld = 6
    
    NEd = abs(joint.column.Ned)
    MEd = joint.column.Med
    h = joint.column.h
    tf = joint.column.tf
    NR = 0.5*NEd + MEd/(h-tf)
    NL = 0.5*NEd - MEd/(h-tf)
    
    print("NR = {0:4.2f}; NL ={1:4.2f}".format(NR,NL))
    
    e1 = 50.0
    p = 180.0
    
    if NL < 0:
        zR = 0.5*(h-tf)
        zL = 0.5*h+e1        
        print("zR = {0:4.2f} mm ; zL ={1:4.2f}".format(zR,zL))

    z = zR+zL    
    print("z = {0:4.2f} mm".format(z))
    
    NR_Ed = NEd*zL/z + MEd/z
    NL_Ed = NEd*zR/z - MEd/z
    
    print("NR_Ed = {0:4.2f} kN".format(NR_Ed*1e-3))
    print("NL_Ed = {0:4.2f} kN".format(NL_Ed*1e-3))
    
    c = joint.c
    
    print("c = {0:4.2f} mm".format(c))
    
    FCRd = joint.FCRd()
    Fc_fc_Rd = joint.Fc_fc_Rd()
    
    print("FCRd = {0:4.2f} kN".format(FCRd*1e-3))
    print("Fc_fc_Rd = {0:4.2f} kN".format(Fc_fc_Rd*1e-3))
    
    m = e1-0.8*math.sqrt(2)*joint.flange_weld
    outstand_top = 110
    d_bolt = 24
    
    Fbolt_Rd = 0.9*550*352/1.25
    print("Fbolt_Rd = {0:4.2f} kN".format(Fbolt_Rd*1e-3))
    
    e = outstand_top + 0.5*d_bolt
    ex = c-e1
    
    leff_cp = min([2*math.pi*m,math.pi*m+p,math.pi*m+2*e])
    leff_nc = min([4*m+1.25*e,e+2*m+0.625*ex,0.5*(joint.column.b+2*c),0.5*p+2*m+0.625*ex])
    
    print("m = {0:4.2f} mm".format(m))
    print(leff_cp)
    print(leff_nc)
    
    leff_1 = min(leff_nc,leff_cp)
    Mpl_1_Rd = 0.25*leff_1*joint.tp**2*joint.fyp
    FT_12_Rd = Mpl_1_Rd/m
    
    print("Mpl_1_Rd = {0:4.2f} kNm".format(Mpl_1_Rd*1e-6))
    print("FT_12_Rd = {0:4.2f} kN".format(FT_12_Rd*1e-3))
    
    F_t_wb_Rd = leff_1*joint.column.tw*joint.column.fy
    print("Ft_wb_Rd = {0:4.2f} kN".format(F_t_wb_Rd*1e-3))
    
    
    
    #joint.fcd
    
    """


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
from metku.optimization.solvers import *
from metku.optimization.problems.truss import TrussProblem
from metku.sections.steel.catalogue import ipe_profiles
from metku.frame2d.frame2d import *

ipe_profiles = list(ipe_profiles.keys())



frame = Frame2D(simple=[3, 2, 3400, 5000], supports='fixed')
beams = []
columns = []
for mem in frame.members.values():
    if mem.mtype == 'beam':
        beams.append(mem)
        frame.add(LineLoad(mem, [-50, -50], 'y'))
    else:
        columns.append(mem)


beam_group = {
    'name': 'Beam Variable',
    'objects': beams,
    'property': 'profile',
    'profiles': ipe_profiles,
    'var_type': 'index'
}

column_group = {
    'name': 'Column Variable',
    'objects': columns,
    'property': 'profile',
    'profiles': ipe_profiles,
    'var_type': 'index'
}

frame.generate()
frame.calculate()


problem = TrussProblem(
    name='Frame Example',
    structure=frame,
    groups=[beam_group, column_group]
)

solver = GA(pop_size=30)
solver.solve(problem, maxiter=50, verb=True)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
import numpy as np
#from frame2d import Frame2D, LineLoad
from scipy.optimize import brute, basinhopping, dual_annealing
from timeit import default_timer as timer

class FrameOptimizer:
    
    def __init__(self,
                 frame,
                 opt_algorithm='brute',
                 lb=None,
                 ub=None,
                 costf=None,
                 constf=None,
                 maxiter=100):
    
        if not frame.is_generated:
            raise TypeError('Frame must be generated before optimization')
        self.frame = frame
        if len(lb):
            self.lb = lb
        else:
            self.lb = [50] * len(frame.members)           
        if len(ub):
            self.ub = ub
        else:
            self.ub = [1000] * len(frame.members)  
        if len(self.lb) != len(self.ub):
            raise ValueError('Lower and upper boundary must be same length')
        self.opt_algorithm = opt_algorithm
        self.costf = costf
        self.constf = constf
        self.maxiter = maxiter
        
      
    def cost_function(self, X, ratio_limit):
        
        R = self.constraint_function(X, ratio_limit)
        R = np.clip(R, -100, 0)
        return self.frame.weight + sum(R**2 * 1e5)
    
    def constraint_function(self, X, ratio_limit=1):
        
        self.change_values(X)
        
        R = []
        for member in self.frame.members.values():
            R.extend(member.r)
            
        return ratio_limit - np.asarray(R)
     
        
    def change_values(self, X):
        for i, member in enumerate(self.frame.members.values()):
            val = round(X[i])
            member.h = val
            #print(member.profile)
        self.frame.calculate()
        
    def optimize(self, ratio_limit=1, debug=True):
        """ Optimizes frame with given optimization algorithm
        
            Returns:
            --------
            :return frame: Optimized frame
            :rtype frame: Frame2D
        """
        
        start = timer()
        
        args = ratio_limit
        if not self.constf:
            constraint_function = self.constraint_function
        else:
            constraint_function = self.constf
        if not self.costf:
            cost_function = self.cost_function
        else:
            cost_function = self.costf
        # initial guess
        x0=[]
        bounds = []
        for i in range(len(self.lb)):
            bounds.append((self.lb[i], self.ub[i]))

        for i in bounds:
            x0.append(np.random.uniform(i[0],i[1]))
        

           
        if debug:
            print("Initial guess x0: ", x0)
            iprint = 2
        else:
            iprint = 0
       
        if self.opt_algorithm == 'brute':
            res = brute(cost_function,
                        bounds,
                        disp=debug)
            print("Optimal values: ", res)
            
        elif self.opt_algorithm == 'basinhopping':
            kwargs = {'args': args}
            res = basinhopping(cost_function,
                                x0,
                                minimizer_kwargs=kwargs,
                                stepsize=30,
                                niter=self.maxiter,
                                interval=10,
                                disp=debug)
            self.change_values(res.x)
            print("Optimal values: ", res.x)
        
        elif self.opt_algorithm == 'dual_annealing':
            pass
        end = timer()
        print(f'Time elapsed: {end - start :.2f} s')
        return self.frame
    


if __name__ == '__main__':
    
    frame = Frame2D(simple=[2,1,5,10], supports='fixed', num_elements=2)
    frame.add(LineLoad(frame.members[4], [-50, -50], 'y'))
    frame.add(LineLoad(frame.members[5], [-50, -50], 'y'))
    #frame.add(LineLoad(frame.members[8], [-50, -50], 'y'))
    #frame.add(LineLoad(frame.members[9], [-50, -50], 'y'))
    frame.hinge_joints()
    frame.generate()
    optimizer = FrameOptimizer(frame,
                               'basinhopping',
                               maxiter=200)
    opt_frame = optimizer.optimize(debug=True, ratio_limit=1)
    opt_frame.plot()
    



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 28 14:06:38 2021

@author: kmela
"""

import numpy as np
import matplotlib.pyplot as plt

class FrameNode:
    """ Class for frame nodes. Used mainly for storing coordinate data """
    
    def __init__(self,coordinates = [0.0,0.0,0.0]):
        """
        

        Parameters
        ----------
        coordinates : list, optional
            nodal coordinates. The default is [0.0,0.0,0.0].

        Returns
        -------
        None.

        """
        
        self.node_id = ""
        self.coords = np.asarray(coordinates)
        self.fem_node = None
        
        self.members = [] # List of members to which the node is connected
        self.symmetry_node = None # Another frame node making a symmetry pair
    
    def __repr__(self):
        
        return f'FrameNode [{self.node_id}]: ({self.x:.3f},{self.y:.3f})'
    
    @property
    def x(self):
        """ x coordinate """
        return self.coords[0]
    
    @x.setter
    def x(self,val):
        """ Set x coordinate value """
        self.coords[0] = val
        # Update also FE node, if it exists
        if not self.fem_node is None:
            self.fem_node.x = val
        
        # Update member lengths
        for mem in self.members:
            mem.member.length = mem.length()
        
    @property
    def y(self):
        """ y coordinate """
        return self.coords[1]
    
    @y.setter
    def y(self,val):
        """ Set y coordinate value """
        self.coords[1] = val
        # Update also FE node, if it exists
        if not self.fem_node is None:
            self.fem_node.y = val
        
        # Update connected member lengths
        for mem in self.members:
            mem.member.length = mem.length()
    
    @property
    def z(self):
        """ z coordinate """
        return self.coords[2]
    
    @z.setter
    def z(self,val):
        """ Set z coordinate value """
        self.coords[2] = val
        
        # Update also FE node, if it exists
        if not self.fem_node is None:
            self.fem_node.z = val
        
        # Update member lengths
        for mem in self.members:
            mem.member.length = mem.length()
    
    def plot(self, print_text=True, marker='ok', axes=None):
        """ Plots the node """
        
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        ax.plot(self.x,self.y,marker)
        
        if print_text:
            ax.text(self.x,self.y,self.node_id)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela

import numpy as np

from ortools.linear_solver import pywraplp

from metku.optimization.solvers.optsolver import OptSolver
from metku.optimization.variables import DiscreteVariable, BinaryVariable

class MISLP(OptSolver):

    def __init__(self, move_limits=(0.1, 0.1), gamma=1e-3, C=2e5, beta=100, move_limit_type='range',only_potential=False):
        super().__init__()
        self.move_limits = np.asarray(move_limits)
        self.move_limits_hist = [np.asarray(move_limits)]
        self.move_limit_type = move_limit_type
        self.gamma = gamma
        self.C = C
        self.beta = beta
        self.milp = None
        self.binary_vars_exist = False
        self.beta_vals = []
        self.milps = []
        self.only_potential = only_potential
        
    def take_action(self):
        """
        Defines action to take
        :return: action
        """
        
        
        # Linearize problem
        #print("X= ",self.X)
        #A, B, df, fx = self.problem.linearize(*self.X.copy())
        print("Linearize...")
        A, B, df, fx = self.problem.linearize(self.X,self.only_potential)
        print("Done")
        # Create CBC solver
        CBC_solver = pywraplp.Solver('MISLP',
                           pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING)

        # Number of constraints
        n = A.shape[0]
        # Number of variables
        m = A.shape[1]

        if self.only_potential:
            act_cons = [con for con in self.problem.cons if con.potential]            
        else:
            act_cons = self.problem.cons

        # Create continuous variables
        x = {}
        # Save discrete variables' indices to map binary variables to
        # these relaxed variables
        discrete_ids = []
        for i, var in enumerate(self.problem.vars):
            #delta = var.ub - var.lb
            #lb, ub = delta * self.move_limits
            
            # The variables in the linearized problem are direction vector
            # components, i.e. d[i] = x[i] - X[i]
            if self.move_limit_type == "range":
                delta = var.ub - var.lb                        
            elif self.move_limit_type == 'relative':
                delta = var.value
            
            lb, ub = self.move_limits * delta
            
            
            
            if isinstance(var, DiscreteVariable):
                discrete_ids.append(i)
                # Find discrete value closest to current variable value
                idx = np.argmin((np.asarray(var.values) - var.value)**2)

                if idx == 0:
                    # The first value is closest
                    ub = max(ub, var.values[1] - var.value)
                elif idx == len(var.values) -1:
                    # The last value is closest
                    lb = max(lb, var.value - var.values[idx - 1])
                else:
                    # Closest value is not the first or last in the list
                    lb = max(lb, var.value - var.values[idx - 1])
                    ub = max(ub, var.values[idx + 1] - var.value)
            
            if isinstance(var,BinaryVariable):
                x[i] = CBC_solver.BoolVar(var.name)
                # If binary variables exist in the problem, assume
                # they are related to linear constraints that
                # control the cross-section of a member. In that case
                # no additional binary variables are created in MISLP
                self.binary_vars_exist = True
            else:                
                lb = max(var.value - lb, var.lb)
                ub = min(var.value + ub, var.ub)
                                
                name = str(var.target['property']) + str(i)
                x[i] = CBC_solver.NumVar(float(lb),float(ub),name)
                    
        
        beta = CBC_solver.NumVar(0, self.beta, 'Beta')
        # Linear Constraints
        # Ax <= B
        
        for i in range(n):
            if act_cons[i].type == '<':                
                CBC_solver.Add(CBC_solver.Sum([A[i, j] * x[j]
                                               for j in range(m)]) <= B[i] + beta)
            elif act_cons[i].type == '=':
                CBC_solver.Add(CBC_solver.Sum([A[i, j] * x[j]
                                               for j in range(m)]) == B[i])
        
        if not self.binary_vars_exist:
            # Create binary variables
            discrete_vars = [var for var in
                             self.problem.vars if isinstance(var, DiscreteVariable)]
            y = {}
            for i, var in enumerate(discrete_vars):
                for j in range(len(var.values)):
                    if (var.id, j) not in y.keys():
                        y[var.id, j] = CBC_solver.BoolVar(f'y{var.id},{j}')
            #print(y)
            # Create binary constraints
            for i, var in enumerate(discrete_vars):
                # Binary constraint
                # sum(bin_vars) == 1            
                CBC_solver.Add(CBC_solver.Sum([y[var.id, j]
                                               for j in range(len(var.values))]) == 1)            
                # Binary property constraint
                # Ai == sum(A[i][bin_idx])
                idx = discrete_ids[i]            
                CBC_solver.Add(CBC_solver.Sum([y[var.id, j] * var.values[j]
                                               for j in range(len(var.values))]) == x[idx])

        # Objective
        CBC_solver.Minimize(CBC_solver.Sum(df[i] * x[i] for i in range(m)) + self.C * beta)

        # Solve
        sol = CBC_solver.Solve()
        self.beta_vals.append(beta.solution_value())
        self.milp = CBC_solver
        self.milps.append(CBC_solver)
        
        if sol == 2:
            print("SOLUTION INFEASIBLE")
            return np.zeros_like(self.X)

        
        # X = [j for i, var in enumerate(discrete_vars)
        #      for j in range(len(var.values)) if y[var.id, j].solution_value() == 1.0]
        #
        #
        # print("MISLP: ", [a.solution_value() for a in x.values()])
        # for i in range(len(X), len(x)):
        #     X.append(x[i].solution_value())
        #
        # print(X)
        # # for i, var in enumerate(discrete_vars):
        # #     print(x[i].solution_value())
        # #     print(var.profiles[X[i]])
        #
        # # for i, var in enumerate(self.problem.vars):
        # #     for j in range(var.ub):
        # #         print(i, j, y[i,j].solution_value())
        #
        # X = np.asarray(X)
        #
        # print(X)
        # for i in discrete_ids:
        #     print(int(X[i]))
        #     X[i] = discrete_vars[i].values[int(X[i])]
        X = np.asarray([round(var.solution_value(), 3) for var in x.values()])
        return X - self.X

    def step(self, action):

        self.move_limits *= (1 - self.gamma)
        self.beta *= (1 - self.gamma)
        self.move_limits_hist.append(list(self.move_limits))
        self.X += action
        # for i in range(len(self.X)):
        #     self.X[i] = np.clip(self.X[i], self.problem.vars[i].lb,
        #                         self.problem.vars[i].ub)

        #self.problem.substitute_variables(self.X)

        return self.X.copy(), 1, False, 'INFO'


def ortools_mip_feasibility(milp,x=None):
    """ Checks feasibility of 'x' for ortools problem 'milp' """
    G_TOL = 1e-5
    
    if x is None:
        x = np.array([var.SolutionValue() for x in milp.variables()])
    else:
        x = np.array(x)
    
    res = True
    
    cons = milp.constraints()
    variables = milp.variables()
    for con in cons:
        a = np.array([con.GetCoefficient(var) for var in variables])
        if a.dot(x) > con.ub() + G_TOL or a.dot(x) < con.lb() -G_TOL:
            print("Infeasible constraint")
            print(con.name())
            res = False
            break

    return res

if __name__ == '__main__':
    from metku.optimization.benchmarks import *
    import matplotlib.pyplot as plt

    ten_bar_DLM = [41, 0, 38, 31, 0, 0, 27, 38, 37, 0]
    # [40  0 38 34  0  0 27 38 38  0] 2497.67
    problem = TenBarTruss(prob_type='discrete')
    solver = MISLP(move_limits=[0.3, 0.3], gamma=1e-2)
    x0 = [var.ub for var in problem.vars]
    solver.solve(problem, x0=x0, maxiter=300, log=False, verb=True)
    problem(solver.X)

    # plt.plot(solver.fvals)
    # plt.show()
    problem.structure.plot()
    #problem(ten_bar_DLM)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
from metku.frame2d.frame2d import *
from metku.optimization.structopt import *
from metku.truss2d import Truss2D
from metku.optimization.solvers import DiscreteVNS


RHS_PROFILES = ['RHS 40X40X2.0', 'RHS 40X40X2.5', 'RHS 50X50X2.0',
                'RHS 40X40X3.0', 'RHS 60X60X2.0', 'RHS 50X50X2.5',
                'RHS 70X70X2.0', 'RHS 50X50X3.0', 'RHS 60X60X2.5',
                'RHS 80X80X2.0', 'RHS 70X70X2.5', 'RHS 60X60X3.0',
                'RHS 50X50X4.0', 'RHS 80X80X2.5', 'RHS 70X70X3.0',
                'RHS 60X60X4.0', 'RHS 90X90X2.5', 'RHS 80X80X3.0',
                'RHS 100X100X2.5', 'RHS 70X70X4.0', 'RHS 90X90X3.0',
                'RHS 60X60X5.0', 'RHS 100X100X3.0', 'RHS 120X120X2.5',
                'RHS 80X80X4.0', 'RHS 70X70X5.0', 'RHS 90X90X4.0',
                'RHS 140X140X2.5', 'RHS 120X120X3.0', 'RHS 80X80X5.0',
                'RHS 150X150X2.5', 'RHS 100X100X4.0', 'RHS 140X140X3.0',
                'RHS 90X90X5.0', 'RHS 150X150X3.0', 'RHS 120X120X4.0',
                'RHS 100X100X5.0', 'RHS 160X160X3.0', 'RHS 140X140X4.0',
                'RHS 100X100X6.0', 'RHS 120X120X5.0', 'RHS 150X150X4.0',
                'RHS 160X160X4.0', 'RHS 140X140X5.0', 'RHS 120X120X6.0',
                'RHS 180X180X4.0', 'RHS 150X150X5.0', 'RHS 120X120X7.1',
                'RHS 160X160X5.0', 'RHS 200X200X4.0', 'RHS 140X140X6.0',
                'RHS 150X150X6.0', 'RHS 120X120X8.0', 'RHS 180X180X5.0',
                'RHS 160X160X6.0', 'RHS 120X120X8.8', 'RHS 180X180X5.6',
                'RHS 200X200X5.0', 'RHS 150X150X7.1', 'RHS 120X120X10.0',
                'RHS 180X180X6.0', 'RHS 160X160X7.1', 'RHS 220X220X5.0',
                'RHS 150X150X8.0', 'RHS 200X200X6.0', 'RHS 160X160X8.0',
                'RHS 150X150X8.8', 'RHS 180X180X7.1', 'RHS 250X250X5.0',
                'RHS 220X220X6.0', 'RHS 160X160X8.8', 'RHS 150X150X10.0',
                'RHS 180X180X8.0', 'RHS 200X200X7.1', 'RHS 160X160X10.0',
                'RHS 180X180X8.8', 'RHS 250X250X6.0', 'RHS 300X300X5.0',
                'RHS 220X220X7.1', 'RHS 200X200X8.0', 'RHS 260X260X6.0',
                'RHS 180X180X10.0', 'RHS 200X200X8.8', 'RHS 220X220X8.0',
                'RHS 250X250X7.1', 'RHS 300X300X6.0', 'RHS 260X260X7.1',
                'RHS 220X220X8.8', 'RHS 200X200X10.0', 'RHS 250X250X8.0',
                'RHS 180X180X12.5', 'RHS 260X260X8.0', 'RHS 220X220X10.0',
                'RHS 300X300X7.1', 'RHS 250X250X8.8', 'RHS 260X260X8.8',
                'RHS 200X200X12.5', 'RHS 300X300X8.0', 'RHS 250X250X10.0',
                'RHS 400X400X6.0', 'RHS 260X260X10.0', 'RHS 300X300X8.8',
                'RHS 400X400X7.1', 'RHS 250X250X12.5', 'RHS 300X300X10.0',
                'RHS 260X260X12.5', 'RHS 400X400X8.0', 'RHS 400X400X8.8',
                'RHS 300X300X12.5', 'RHS 400X400X10.0', 'RHS 400X400X12.5']



class KTruss24m(OptimizationProblem):
    """
    General class for creating optimization problems that utilize the Frame2D -module

    Finite element analysis is done in the objective function and it's should always
    be calculated before constraints.

    """
    def __init__(self ,name="Test Truss", profiles=RHS_PROFILES):
        super().__init__(name)

        self.profiles = profiles
        self._create_structure()
        self._create_variables()
        self.create_objective()
        self.create_constraints()



    def _create_structure(self):

        truss = Truss2D(simple={'H0': 0,
                                'H1': 1200,
                                'H2': 2400,
                                'L1': 12000,
                                'n': 18})


        truss.add(FixedSupport([0, 1200]))
        truss.add(FixedSupport([24000, 1200]))
        truss.add(LineLoad(truss.top_chords[0], [-23.52, -23.52], 'y'))
        truss.add(LineLoad(truss.top_chords[1], [-23.52, -23.52], 'y'))
        truss.generate()
        truss.calculate()
        self.structure = truss
        truss.plot_normal_force()


    def _create_variables(self):

        # Initialize variables as an empty array
        self.vars = []
        # Create profile variables
        # TODO: symmetry, grouping
        for i, mem in enumerate(self.structure.members.values()):
            name = mem.mem_id
            var = DiscreteVariable(name,
                                profiles=self.profiles,
                                target={"property": "PROFILE",
                                        "objects": [mem]})
            self.vars.append(var)

        # Create geometry variables
        # TODO

    def create_objective(self):

        def objective(X):

            weight = 0
            for x, mem in zip(X, self.structure.members.values()):
                weight += mem.weight
            return weight

        self.obj = objective

    def _create_normal_force_constraints(self):
        """
        Creates normal force constraints
        """
        for i, mem in enumerate(self.structure.members.values()):
            def con_fun(X, i=i):
                return mem.ned / mem.NRd - 1
            constr = NonLinearConstraint(con_fun, name='Normal Force ' + str( i +1))
            self.cons.append(constr)


    def _create_shear_force_constraints(self):
        """
        Creates shear force constraints
        """
        for i, mem in enumerate(self.structure.members.values()):
            def con_fun(X, i=i):
                return mem.ved / mem.VRd - 1
            constr = NonLinearConstraint(con_fun, name='Shear Force ' + str( i +1))
            self.cons.append(constr)


    def _create_bending_moment_constraints(self):
        """
        Creates bending moment constraints about y-axis
        """
        for i, mem in enumerate(self.structure.members.values()):
            def con_fun(X, i=i):
                return mem.med / mem.MRd[0] - 1
            constr = NonLinearConstraint(con_fun, name='Bending Moment ' + str( i +1))
            self.cons.append(constr)

    def create_constraints(self, N=True, V=True, M=True):

        # Initialize constraints as an empty list
        self.cons = []
        # Create normal force constraints
        if N:
            self._create_normal_force_constraints()
        if V:
            self._create_shear_force_constraints()
        if M:
            self._create_bending_moment_constraints()

if __name__ == '__main__':

    truss = KTruss24m()
    solver = DiscreteVNS()
    #solver.solve(truss, maxtime=60)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
from metku.frame2d.member import *
from metku.frame2d.loads import *
from metku.frame2d.supports import *

import matplotlib.pyplot as plt
import metku.framefem.framefem as ff


class Frame:
    """
    CLASS DEFINITION
    """

    def __init__(self):

        # Initialize empty dicts
        self.members = {}
        self.loads = {}
        self.supports = {}
        self.loadcases = {}


    def generate_fem_model(self, load_id, supp_id):
        """
        Generates new fem model
        :param load_id: loads to include
        :param supp_id: supports to include
        :return: generated frame
        """
        # Initialize frame
        fem_model = ff.FrameFEM()
        # Members
        for mem in self.members.values():

            # for node in mem.nodes.values():
            #     fem_model.add_node(*node.coord)

            for element in mem.elements.values():
                fem_model.add(element)
        # Supports
        for support in self.supports.values():
            if support.sid == supp_id:
                fem_model.add(support)
        # Loads
        for load in self.loads.values():
            if load.load_id == load_id:
                if isinstance(load, PointLoad):
                    pl = ff.PointLoad(load.load_id,
                                      load.node,
                                      load.F,
                                      load.scale)
                    fem_model.add_load(pl)
                elif isinstance(load, LineLoad):
                    pass

        return fem_model

    def add(self, obj, intersect=True):
        """
        Adds object to frame
        Object can be Member, Load, Support or Truss

        :param obj: obj to be added to frame
        """
        if isinstance(obj, FrameMember):
            self.add_member(obj, intersect)

        elif isinstance(obj, PointLoad):
            self.add_pointload(obj)

    def add_member(self, member, intersect=True):
        """
        Adds member to frame
        Adds nodes between intersecting members
        :param member: FrameMember object
        """
        for mem in self.members.values():
            point = member.intersection(mem)
            if len(point) and intersect:
                if list(point) in mem.nodal_coordinates:

                    for nid, node in mem.nodes.items():
                        if np.all(node.coord == point):
                            member.intersecting_nodes[nid] = node
                elif list(point) in member.nodal_coordinates:
                    for nid, node in member.nodes.items():
                        if np.all(node.coord == point):
                            mem.intersecting_nodes[node.nid] = node
                else:
                    nid = member.mem_id * 100 + len(member.nodes)
                    node = FEMNode(nid, *point)
                    mem.intersecting_nodes[nid] = node
                    member.intersecting_nodes[nid] = node
                mem.nels += 1
                member.nels += 1
                mem.update()
                member.update()

        self.members[member.mem_id] = member

    def add_pointload(self, pl):
        """
        Adds point load to frame
        :param pl: PointLoad object
        """
        for mem in self.members.values():
            point = mem.intersection(pl._coord)
            if len(point):
                if list(point) in mem.nodal_coordinates:
                    node = [node for node in mem.nodes.values() if np.all(node.coord == point)][0]
                    pl.node = node
                else:
                    nid = mem.mem_id * 100 + len(mem.nodes)
                    node = FEMNode(nid, *point)
                    mem.nodes[nid] = node
                    pl.node = node



    def calculate(self, load_id=0, supp_id=0):
        """
        Runs linear finite element analysis
        :param load_id: load id
        """
        fem_model = self.generate_fem_model(load_id, supp_id)
        fem_model.add_loadcase(supp_id=supp_id, load_id=load_id)
        fem_model.nodal_dofs()
        fem_model.linear_statics()


    def plot(self, show=True):
        """
        Plots the frame
        :param show: boolean
        """

        for mem in self.members.values():
            mem.plot(show=False)
        if show:
            plt.show()


if __name__ == '__main__':
    frame = Frame()
    mem = FrameMember([0, 0], [1000, 1000], num_elements=1)
    mem2 = FrameMember([-100, 750], [1000, 500], num_elements=1)
    mem3 = FrameMember([100, 100], [-100, 1000], num_elements=3)
    frame.add(mem)
    frame.add(mem2)
    frame.add(mem3)
    for mem in frame.members.values():
        mem.plot_elements(False)
    plt.show()
    fem = frame.generate_fem_model(0,0)
    fem.draw()



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Wed Mar 14 23:12:31 2018

Hot-rolled I- and H-section

@author: kmela
"""
import math

from metku.eurocodes.en1993 import en1993_1_1, constants
from metku.sections.steel.steel_section import SteelSection
from metku.sections.steel.catalogue import profile

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.lines as mlines
import matplotlib.path as mpath
import numpy as np

class ISection(SteelSection):
    """ Base class for I -sections """

    def __init__(self, H, B, tf, tw, r, fy=355):
        """
            H -- height
            B -- width
            T -- wall thickness
            fy -- yield strength
        """
        self.h = H
        self.b = B
        self.tf = tf
        self.tw = tw
        self.r = r
        
        #self.fy = fy

        A, Au, Ashear, I, It, Iw, Wpl, Wel = cross_section_properties(B, H, tf, tw, r)
        self.Iw = Iw
        self.It = It
        super().__init__(fy, A, I, Au, Wpl, Wel, Ashear)

        """ Determine buckling curve: EN 1993-1-1, Table 6.2 """
        if H / B > 1.2:
            if tf <= 40:
                if fy <= 420:
                    buck_y = "a"
                    buck_z = "b"
                else:
                    buck_y = "a0"
                    buck_z = "a0"
            else:
                if fy <= 420:
                    buck_y = "b"
                    buck_z = "c"
                else:
                    buck_y = "a"
                    buck_z = "a"
        else:
            if tf <= 100:
                if fy <= 420:
                    buck_y = "b"
                    buck_z = "c"
                else:
                    buck_y = "a"
                    buck_z = "a"
            else:
                if fy <= 420:
                    buck_y = "d"
                    buck_z = "d"
                else:
                    buck_y = "c"
                    buck_z = "c"

        self.imp_factor = [en1993_1_1.buckling_curve[buck_y],
                           en1993_1_1.buckling_curve[buck_z]]
        
        
        if H/B <= 2.0:
            self.imp_factor_LT_gen = en1993_1_1.buckling_curve["a"]
            self.imp_factor_LT = en1993_1_1.buckling_curve["b"]
        else:
            self.imp_factor_LT_gen = en1993_1_1.buckling_curve["b"]
            self.imp_factor_LT = en1993_1_1.buckling_curve["c"]


    def __repr__(self):
        return f"{type(self).__name__} {self.h:.0f}"
    
    def robot(self):
        # Format name of the profile according to Robot Structural Analysis
        return self.__repr__()
    
    def info(self,latex=False):
        """ Prints a bunch of section properties """
        
        if latex:
            print("*** " + self.__repr__() + " ***")
            print(" $h = {0:.0f}{1:s}$".format(self.h,'\\,\\unit{mm}'))
            print(" $t_w = {0:.1f}{1:s}$".format(self.tw,'\\,\\unit{mm}'))
            print(" $b = {0:.0f}{1:s}$".format(self.b,'\\,\\unit{mm}'))
            print(" $t_f = {0:.1f}{1:s}$".format(self.tf,'\\,\\unit{mm}'))
            print(" $r = {0:.0f}{1:s}$".format(self.r,'\\,\\unit{mm}'))
            print(" $A = {0:.2f}{1:s}$".format(self.A*1e-2,'\\ee{2}\\,\\squnit{mm}'))
            print(" $I_y = {0:.2f}{1:s}$".format(self.I[0]*1e-4,'\\ee{4}\\,\\quunit{mm}'))
            print(" $I_z = {0:.2f}{1:s}$".format(self.I[1]*1e-4,'\\ee{4}\\,\\quunit{mm}'))
            print(" $W_el,y = {0:.2f}{1:s}$".format(self.Wel[0]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))
            print(" $W_el,z = {0:.2f}{1:s}$".format(self.Wel[1]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))
            print(" $W_pl,y = {0:.2f}{1:s}$".format(self.Wpl[0]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))
            print(" $W_pl,z = {0:.2f}{1:s}$".format(self.Wpl[1]*1e-3,'\\ee{3}\\,\\cuunit{mm}'))
            print(" $I_t = {0:.2f}{1:s}$".format(self.It*1e-4,'\\ee{4}\\,\\quunit{mm}'))
            print(" $I_w = {0:.2f}{1:s}$".format(self.Iw*1e-12,'\\ee{12}\\,\\text{mm}^{6}'))
        else:
            print("*** " + self.__repr__() + " ***")
            print(" h = {0:.0f} mm".format(self.h))
            print(" tw = {0:.1f} mm".format(self.tw))
            print(" b = {0:.0f} mm".format(self.b))
            print(" tf = {0:.1f} mm".format(self.tf))
            print(" r = {0:.0f} mm".format(self.r))
            print(" A = {0:.2f} (100 mm2)".format(self.A*1e-2))
            print(" Iy = {0:.2f} (10^4 mm4)".format(self.I[0]*1e-4))
            print(" Iz = {0:.2f} (10^4 mm4)".format(self.I[1]*1e-4))
            print(" Wel,y = {0:.2f} (10^3 mm4)".format(self.Wel[0]*1e-3))
            print(" Wel,z = {0:.2f} (10^3 mm4)".format(self.Wel[1]*1e-3))
            print(" Wpl,y = {0:.2f} (10^3 mm4)".format(self.Wpl[0]*1e-3))
            print(" Wpl,z = {0:.2f} (10^3 mm4)".format(self.Wpl[1]*1e-3))
            print(" It = {0:.2f} (10^4 mm4)".format(self.It*1e-4))
            print(" Iw = {0:.2f} (10^12 mm6)".format(self.Iw*1e-12))
        

    @property
    def hw(self):
        """ Straight part of the web """
        return self.h - 2 * self.tf - 2 * self.r

    @property
    def cf(self):
        """ Straight part of the flange """
        return 0.5 * (self.b - self.tw) - self.r
    
    @property
    def box_section_factor(self):
        """ Section factor for fire, considering the profile
            as a box
        """
        return 2*(self.b+self.h)/self.A*1e3
    
    
    def shadow_effect(self):
        """ Shadow effect factor ksh, from EN 1993-1-2 """
        
        return 0.9*self.box_section_factor/self.section_factor

    def flange_class(self,verb=False):
        """ Determine class of compressed flange """
        # cf = 0.5*(self.b - self.tw) - self.r
        rf = self.cf / self.tf
        cFlange = en1993_1_1.outstand_part_in_compression(rf, self.eps)

        if verb:
            print("Flange classification (outstand element):")
            print("cf = {0:4.2f}, tf = {1:4.2f}".format(self.cf,self.tf))
            print("cf/tf = {0:4.2f}".format(rf))
            print("Flange class = {0}".format(cFlange))
            

        return cFlange
    
    def compressed_flange(self):
        """ Returns cross-sectional area and second moment of
            area of the compression flange which include one flange
            and 1/3 of the height of the compressed web.
            It is assumed that the cross-section is in bending, so half
            of the web is in compression.
        """
        h0 = self.h-2*self.tf
        Af = 0.5*self.A-1/3*h0*self.tw
        
        hw = self.hw
        # Second moment of area of one flange and corners
        Ifz0 = 0.5*self.Iz-0.5/12*hw*self.tw**3
        R = self.r
        hFlat = 1/6*h0-R
        Ifz = Ifz0 + 1/12*hFlat*self.tw**3
        
        return Af, Ifz
        
        

    def web_class_comp(self,verb=False):
        """ Determine class of compressed web """
        rw = self.hw / self.tw
        cWeb = en1993_1_1.internal_part_in_compression(rw, self.eps)

        if verb:
            print("Web classification (internal part in compression):")
            print("hw = {0:4.2f}, tw = {1:4.2f}".format(self.hw,self.tw))
            print("hw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def web_class_bend(self,verb=False):
        """ Determine class of web in bending """
        rw = self.hw / self.tw
        cWeb = en1993_1_1.internal_part_in_bending(rw, self.eps)

        if verb:
            print("Web classification (internal part in bending):")
            print("hw = {0:4.2f}, tw = {1:4.2f}".format(self.hw,self.tw))
            print("hw/tw = {0:4.2f}".format(rw))
            print("Web class = {0}".format(cWeb))

        return cWeb

    def web_class_comp_bend(self, Ned, verb=False, approach = 1):
        """ Determine class of web in combined bending and compression 
            
            Approach:
                Take Ned as given and fix it. Then see how large the
                bending moment can be such that the entire section
                has been plastified.
                
            Approach 2:
                Increase NEd and MEd by the same factor until
                plastification of the cross-section has been reached,
                
                or until the first yielding begins
        """
        # cw = self.h-2*self.tf-2*self.r
        
        if verb:
            print("Web classification (internal part in compression and bending):")
            print("hw = {0:4.2f}, tw = {1:4.2f}".format(self.hw,self.tw))
        
        """ Calculate stresses for class 3 check """
        sN = self.sigmaN()
        sM = self.sigmaM() 
    
        cw = self.hw
        tw = self.tw
        rw = cw / tw
                
    
        if approach == 2:
            zeta = self.Ned/abs(self.Med)
            Wpl0 = self.Wpl[0] - 0.25*cw*tw
            K = cw*zeta
            
            if abs(zeta) < cw*tw/Wpl0:
                if verb:
                    print("Neutral axis in the web.")
                    print("zeta = {0:4.3f}".format(zeta))                    
                    print(1/K)
                a = 0.5 + 1/K - 0.5/K*np.sqrt(4+K**2 + 4*zeta**2*Wpl0/tw)
            else:
                """ Neutral axis is not in the web """
                print("Neutral axis is not in the web.")
                a = 0.0
            
            """ Stress ratio for class 3 check """
            
            if verb:
                print("sN = {0:4.3f} MPa".format(sN))
                print("sM = {0:4.3f} MPa".format(sM))
            
            if sN > cw/self.h*abs(sM): # and sN > -cw/self.h*sM:
                """ Only positive stresses in the web """
                p = 1
            else:    
                """ Both postiive and negative stresses in the web:
                    calculate stress ratio.
                    Here, the calculation is done with respect to the
                    extreme fibres of the cross-section. In SEMI-COMP+,
                    the ratio is calculated for the stresses in the
                    straight part of the web.
                """
                p = (sN+sM)/(sN-sM)
            
            #z0 = zeta*self.I[0]/self.A
            #p = -(0.5*self.h-z0)/(0.5*self.h+z0)
            
            if verb:
                print("hw/tw = {0:4.2f}".format(rw))
                print("NEd/MEd = zeta = {0:4.3f}".format(zeta))
                print("alpha = {0:4.3f}".format(a))
                print("psi = {0:4.3f}".format(p))
            
        if approach == 1:
            NRweb = cw*tw*self.fy
            if -self.Ned >= NRweb:
                a = 1.0
            elif -self.Ned <= -NRweb:
                a = 0.0
            else:
                a = 0.5*(1-self.Ned/NRweb)
                
            """ Stress ratio for class 3 check """
            
            if verb:
                print("sN = {0:4.3f} MPa".format(sN))
                print("sM = {0:4.3f} MPa".format(sM))
            
            if sN > cw/self.h*abs(sM): # and sN > -cw/self.h*sM:
                """ Only positive stresses in the web """
                p = 1
            else:    
                """ Both postiive and negative stresses in the web:
                    calculate stress ratio.
                    Here, the calculation is done with respect to the
                    extreme fibres of the cross-section. In SEMI-COMP+,
                    the ratio is calculated for the stresses in the
                    straight part of the web.
                """
                p = (sN+sM)/(sN-sM)
                
            if verb:                
                print("NEd = {0:4.3f} kN".format(self.Ned*1e-3))
                print("NR_web = {0:4.3f} kN".format(NRweb*1e-3))
                print("alpha = {0:4.3f}".format(a))
                print("psi = {0:4.3f}".format(p))
            
        #Ac = 0.5 * (self.A + Ned / (self.fy / constants.gammaM0))
        #a = Ac / self.A
        #p = -1        
        cWeb = en1993_1_1.internal_part_comp_bend(rw, self.eps, a, p)
        
        if verb:
            print("Web class = {0}".format(cWeb))

        return cWeb

    def moment_axial_force_interact(self, UN, MRd):
        """ Interaction rule for section resistance for combined
            axial force and bending
            
            input: UN .. NEd/NRd
                  MRd .. moment resistance                 
        """
        aw = min((self.A - 2 * self.b * self.tf) / self.A, 0.5)

        if UN > 1.0:
            MNRd = 0.0
        else:
            MNRd = MRd * min((1 - UN) / (1 - 0.5 * aw), 1)

        return MNRd

    

    def draw(self,axes=None,origin=[0,0],theta=0):
        """ Draw the profile 
            input:
                axes .. matplotlib.axes.Axes object. If this is given,
                        the profile will be drawn to that Axes
                theta .. rotation (optional)
        """
        
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        if theta != 0:
            phi = np.radians(theta)
            c, s = np.cos(phi), np.sin(phi)
            R = np.array(((c,-s), (s, c)))
        
        lw = 1.5
        
        h = self.h
        b = self.b
        tf = self.tf
        tw = self.tw
        r = self.r
        hw = self.hw
                
        xdata = [-0.5*tw-r,-0.5*b,-0.5*b,0.5*b,0.5*b,0.5*tw+r]
        ydata = [-0.5*h+tf, -0.5*h+tf, -0.5*h,-0.5*h,-0.5*h+tf,-0.5*h+tf]
        
        if origin[0] != 0:
            xdata = [x+origin[0] for x in xdata]
        
        if origin[1] != 0:
            ydata = [y+origin[1] for y in ydata]
        
        if theta != 0:
            xrot = [R.dot(np.array([x,y])) for x, y in zip(xdata,ydata)]
            xdata = [x[0] for x in xrot]
            ydata = [x[1] for x in xrot]
        
        bflange = mlines.Line2D(xdata,ydata,linewidth=lw,color='k')
                
        ax.add_line(bflange)
        
        # Bottom chamfers
        if theta != 0:
            x0 = tuple(R.dot(np.array([0.5*tw+r,-0.5*self.hw])))
        else:
            x0 = (0.5*tw+r+origin[0],-0.5*self.hw+origin[1])
                
        x1 = (-(0.5*tw+r)+origin[0],x0[1])        
        
        
        br_corner = patches.Arc(x0,2*r,2*r,0,180+theta,270+theta,linewidth=lw,zorder=4,color='k')
        bl_corner = patches.Arc(x1,2*r,2*r,0,270+theta,360+theta,linewidth=lw,zorder=4,color='k')
        
        ax.add_patch(br_corner)
        ax.add_patch(bl_corner)
        
    
        #ax.plot(0.5*tw+r,-0.5*self.hw,'bo')
        
        top_x_data = xdata
        top_y_data = [0.5*h-tf, 0.5*h-tf, 0.5*h,0.5*h,0.5*h-tf,0.5*h-tf]        
        
        if origin[1] != 0:
            top_y_data = [y+origin[1] for y in top_y_data]
        
        tflange =  mlines.Line2D(top_x_data,top_y_data,linewidth=lw,color='k')
        ax.add_line(tflange)
        
        
        
        # Top chamfers
        tr_corner = patches.Arc((0.5*tw+r+origin[0],0.5*hw+origin[1]),2*r,2*r,0,90,180,zorder=4,color='k',linewidth=lw)
        tl_corner = patches.Arc((-0.5*tw-r+origin[0],0.5*hw+origin[1]),2*r,2*r,0,0,90,zorder=4,color='k',linewidth=lw)
        
        ax.add_patch(tr_corner)
        ax.add_patch(tl_corner)
        
        # Draw web
        xweb_l = [-0.5*tw+origin[0],-0.5*tw+origin[0]]
        yweb = [-0.5*hw+origin[1],0.5*hw+origin[1]]
        
        if theta != 0:
            web_rot = [R.dot(np.array([x,y])) for x, y in zip(xweb_l,yweb)]
            xweb_l = [x[0] for x in web_rot]
            yweb_l = [x[1] for x in web_rot]
        else:
            yweb_l = yweb
        
        lweb = mlines.Line2D(xweb_l,yweb_l,linewidth=lw,color='k')
        ax.add_line(lweb)
        
        xweb_r = [0.5*tw+origin[0],0.5*tw+origin[0]]
        
        rweb = mlines.Line2D(xweb_r,yweb,linewidth=lw,color='k')
        ax.add_line(rweb)
        
        """
        verts = [(0,0),
                 (1,0),
                 (0,1)]
        
        codes = [mpath.Path.]
        
        Iprof_path = mpath.Path(verts,codes)
        
        
        Iprof = patches.PathPatch(Iprof_path,facecolor="none",lw=1.5)
        ax.add_patch(Iprof)
        """
        
        #ax.add_patch(bot_flange)
        #ax.add_patch(web)
        #ax.add_patch(top_flange)
        #ax.set_xlim(-0.5*b, 0.5*b)
        #ax.set_ylim(-0.5*h, 0.5*h)
        
        if axes is  None:
            ax.set_xlim(-b, b)
            ax.set_ylim(-h, h)
        
            ax.set_aspect('equal')
        
        return ax
            

class IPE(ISection):
    """ IPE sections
        Subclass of ISection
    """

    def __init__(self, height=100, fy=355, catalogue=True):
        name = 'IPE ' + str(height)        
        # H,B,tf,tb,r,fy=355
        if catalogue or name in profile.keys():
            try:
                H = profile[name]['h']
                B = profile[name]['b']
                tf = profile[name]['t_f']
                tw = profile[name]['t_w']
                r = profile[name]['r']
            except KeyError:
                print('Error: No profile named ' + name + ' in the catalogue.')
        else:
            H = height
            K = [0.0256, 3.2831, 0.0155, 2.4921, 0.3486, 30.178, 0.0155, 2.4921]
            tf = K[0] * H + K[1]
            tw = K[2] * H + K[3]
            B = K[4] * H + K[5]
            r = K[6] * H + K[7]

        ISection.__init__(self, H, B, tf, tw, r, fy=fy)


class HEA(ISection):
    """ European wide flange sections
        Subclass of ISection        
    """

    def __init__(self, height=100, fy=355, catalogue=True):        
        name = 'HE ' + str(height) + ' ' + 'A'
        # HEIGHT =! TO PROFILE NUMBER

        self.name = name
        # H,B,tf,tb,r,fy=355
        if catalogue:
            H = profile[name]['h']
            B = profile[name]['b']
            tf = profile[name]['t_f']
            tw = profile[name]['t_w']
            r = profile[name]['r']
        else:
            H = height
            K = [0.0294, 5.7651, 0.014, 4.2949, 1.0323, 2.5368, 0.0781, 2.9206]
            tf = K[0] * H + K[1]
            tw = K[2] * H + K[3]

            if H <= 270:
                B = K[4] * H + K[5]
                r = K[6] * H + K[7]
            else:
                B = 300
                if H <= 690:
                    r = 27
                else:
                    r = 30    

        ISection.__init__(self, H, B, tf, tw, r, fy=fy)

    def __repr__(self):
        return self.name
        #return "HEA " + str(self.b-(self.b%10))
    
    def robot(self):
        # Format name suitable for Robot Structural Analysis
        return self.name[:2] + self.name[-1] + self.name[2:-2]

class HEAA(ISection):
    """ European wide flange sections
        Subclass of ISection        
    """

    def __init__(self, height=100, fy=355, catalogue=True):
        name = 'HE ' + str(height) + ' ' + 'AA'
        # HEIGHT =! TO PROFILE NUMBER

        self.name = name
        # H,B,tf,tb,r,fy=355
        if catalogue:
            H = profile[name]['h']
            B = profile[name]['b']
            tf = profile[name]['t_f']
            tw = profile[name]['t_w']
            r = profile[name]['r']
        else:
            H = height
            K = [0.0179, 4.9342, 0.0144, 3.2122, 1.0403, 6.2902, 0.0786, 3.221]
            tf = K[0] * H + K[1]
            tw = K[2] * H + K[3]

            if H <= 264:
                B = K[4] * H + K[5]
                r = K[6] * H + K[7]
            else:
                B = 300
                if H <= 670:
                    r = 27
                else:
                    r = 30

        ISection.__init__(self, H, B, tf, tw, r, fy=fy)


class HEB(ISection):
    """ European wide flange sections
        Subclass of ISection        
    """
    def __init__(self,height=100,fy=355,catalogue=True):
        name = 'HE ' + str(height) + ' ' + 'B'
        self.name = name
        # H,B,tf,tb,r,fy=355
        if name in profile.keys():
            H=profile[name]['h']
            B=profile[name]['b']
            tf = profile[name]['t_f']
            tw = profile[name]['t_w']
            r = profile[name]['r']
        else:
            H = height
            K = [0.0308, 9.5736, 0.0147, 6.2014, 1, 0, 0.0755, 2.7636]
            tf = K[0] * H + K[1]
            tw = K[2] * H + K[3]

            if H <= 280:
                B = K[4] * H + K[5]
                r = K[6] * H + K[7]
            else:
                B = 300
                if H <= 700:
                    r = 27
                else:
                    r = 30

        ISection.__init__(self, H, B, tf, tw, r, fy=fy)

    def __repr__(self):
        return self.name
        #return "HEA " + str(self.b-(self.b%10))
    
    def robot(self):
        # Format name suitable for Robot Structural Analysis
        return self.name[:2] + self.name[-1] + self.name[2:-2]

class HEC(ISection):
    """ HEC sections
        Subclass of ISection
    """

    def __init__(self, height=100, fy=355, catalogue=True):
        name = 'HE ' + str(height) + ' ' + 'C'
        self.name = name
        # H,B,tf,tb,r,fy=355
        if catalogue:
            H = profile[name]['h']
            B = profile[name]['b']
            tf = profile[name]['t_f']
            tw = profile[name]['t_w']
            r = profile[name]['r']
        else:
            H = height
            K = [0.0669, 6.7987, 0.0317, 5.4267, 0.9256, 5.593, 0.0747, 2.0905]
            tf = K[0] * H + K[1]
            tw = K[2] * H + K[3]
            B = K[4] * H + K[5]
            r = K[6] * H + K[7]

        ISection.__init__(self, H, B, tf, tw, r, fy=fy)


class HEM(ISection):
    """ European wide flange sections
        Subclass of ISection        
    """

    def __init__(self, height=100, fy=355, catalogue=True):
        name = 'HE ' + str(height) + ' ' + 'M'
        # HEIGHT =! TO PROFILE NUMBER
        self.name = name
        # H,B,tf,tb,r,fy=355
        if catalogue:
            H = profile[name]['h']
            B = profile[name]['b']
            tf = profile[name]['t_f']
            tw = profile[name]['t_w']
            r = profile[name]['r']
        else:
            H = height
            K = [0.0824, 8.5346, 0.0392, 6.8259, 0.9344, -2.8966, 0.0732, 1.5707]

            if H <= 340:
                tf = K[0] * H + K[1]
                tw = K[2] * H + K[3]
                B = K[4] * H + K[5]
                r = K[6] * H + K[7]
            else:
                tf = 40
                tw = 21
                B = -0.0112 * H + 312.41
                if H <= 716:
                    r = 27
                else:
                    r = 30

        ISection.__init__(self, H, B, tf, tw, r, fy=fy)


class CustomISection(ISection):
    def __init__(self, H, B, tf, tw, r, fy=355):
        self.H = H
        self.B = B
        self.tf = tf
        self.tw = tw
        self.r = r

        ISection.__init__(self, H, B, tf, tw, r, fy=fy)

    def cross_section_properties(self):
        args = cross_section_properties(self.B, self.H, self.tf, self.tw, self.r)
        self.A, self.Au, self.Ashear, self.I, self.It, self.Iw, self.Wpl, self.Wel = args


def cross_section_properties(b, h, tf, tw, r):
    """ Cross-section properties        
        Calculation according to http://sections.arcelormittal.com/fileadmin/redaction/4-Library/
        1-Sales_programme_Brochures/Sales_programme/Sections_MB-ArcelorMittal_FR_EN_DE-V2017-3.pdf        
        page 193 ->
    """
    # Area of section [mm^2]
    A = 2.0 * tf * b + (h - 2.0 * tf) * tw + (4.0 - math.pi) * r ** 2.0

    # Perimeter of cross-section
    Au = 4 * (b - 2 * r) + 2 * (h - 2 * tw) + 2 * math.pi * r

    # Shear area [mm^2]    
    Ashear = A - 2.0 * b * tf + (tw + 2.0 * r) * tf

    # Second moment of area, I_y is stronger direction [mm^4]
    I = [0.0, 0.0]
    I[0] = 1.0 / 12.0 * (b * h ** 3.0 - (b - tw) * (h - 2.0 * tf) ** 3.0) + 0.03 * r ** 4 + 0.2146 * r ** 2.0 * (
                                                                                                                h - 2.0 * tf - 0.4468 * r) ** 2.0
    I[1] = 1.0 / 12.0 * (2.0 * tf * b ** 3.0 + (h - 2.0 * tf) * tw ** 3.0) + 0.03 * r ** 4.0 + 0.2146 * r ** 2.0 * (
                                                                                                                   tw + 0.4468 * r) ** 2.0

    # Torsional constant [mm^4]
    # Taken from Arcelor Mittal, with source SCI P385 Design of Members in Torsion
    #It = 2.0 / 3.0 * (b - 0.63 * tf) *tf ** 3.0 + 1.0 / 3.0 * (h - 2.0 * tf) * tw ** 3.0 + 2.0 * (tw / tf) * (0.145 + 0.1 * r / tf) * (((r + tw / 2.0) ** 2.0 + (r + tf) ** 2.0 - r ** 2.0) / (2.0 * r + tf)) ** 4
    a1 = -0.042+0.2204*tw/tf + 0.1355*r/tf - 0.0865*r*tw/tf**2 - 0.0725*tw**2/tf**2
    D1 = ((tf+r)**2+(r+0.25*tw)*tw)/(2*r+tf)
    It = 2.0/3.0*b*tf**3 + (h-2*tf)*tw**3/3.0 + 2*a1*D1**4 - 0.420*tf**4
    
    """
    It = 2.0/3.0*(b - 0.63*tf)*tf**3.0 + \
         + 1.0/3.0*(h-2.0*tf)*tw**3.0 + \
         + (0.042 + 0.2204*(tw / tf)+0.1355*(r/tf)-0.0865*(r*tw/tf**2) -0.0725*(tw/tf)**2) * (((r + tw/2.0)**2.0 + (r + tf)**2.0 - r**2.0)/(2.0*r + tf))**4.0
    """
    #print(tf,tw,b,h,It)

    # Warping constant [mm^6]
    Iw = (tf * b ** 3.0) / 24.0 * (h - tf) ** 2.0

    # Plastic section modulus [mm^3]
    Wpl = [0.0, 0.0]
    Wpl[0] = (tw * h ** 2.0) / 4.0 + (b - tw) * (h - tf) * tf + (4.0 - math.pi) / 2.0 * r ** 2.0 * (h - 2.0 * tf) + (
                                                                                                                    3.0 * math.pi - 10.0) / 3.0 * r ** 3.0
    Wpl[1] = (b ** 2.0 * tf) / 2.0 + (h - 2.0 * tf) / 4.0 * tw ** 2.0 + r ** 3.0 * (10.0 / 3.0 - math.pi) + (
                                                                                                            2.0 - math.pi / 2.0) * tw * r ** 2.0

    # Elastic section modulus [mm^3]
    Wel = [0.0, 0.0]
    Wel[0] = I[0] / (0.5 * h)
    Wel[1] = I[1] / (0.5 * b)

    # Mass per unit length [kg/mm]
    # g = A*rho

    # Variable epsilon used in design
    # eps = math.sqrt(235.0/fy)

    return A, Au, Ashear, I, It, Iw, Wpl, Wel


if __name__ == '__main__':
    
    p = IPE(400)
    #p.info(latex=True)
    #p.draw(theta = 0)
    
    Npl = p.NRd
    Mpl = p.plastic_bending_resistance()
    NEd = 550e3 #0.17*Npl
    
    UN = NEd/Npl
    MRd = Mpl
    MNRd = p.moment_axial_force_interact(UN, MRd)
    
    p.Ned = -NEd
    p.Med = 350e6 # 0.75*Mpl
    
    UM = p.Med/MRd
    
    print('UN + UM = ',UN+UM)
    
    p.web_class_comp_bend(approach=1,verb=True)


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
@author: Victoria
Pilarin laskentatehtävä.
"""

import time


from metku.frame2d.frame2d import *
from metku.optimization.solvers.optsolver import OptSolver
# from metku.optimization.structopt import *
import metku.optimization.structopt as sopt
from metku.optimization.problems.wi_column import WIColumn
from metku.optimization.solvers import slsqp, slp, slqp, mislp, two_phase
from metku.optimization.solvers.trust_region import TrustRegionConstr
from metku.optimization.solvers.bnb import BnB
from metku.optimization.solvers.lp import LP
from metku.optimization.result_exporter import *
from metku.eurocodes.en1993 import en1993_1_1



class ColumnCalculation(WIColumn):
    # def ColumnCalculation():

    # Määritellään kuormat
    q_wind = 0.00065  # N/mm2
    q_snow = 0.002  # N/mm2
    g_truss = 1  # N/mm
    g_roof = 0.0005  # N/mm2

    # Ristikon jännevälit
    L_list = [24000, 30000, 36000]  # mm
    # L_list = [30000]  # mm
    # Pilarin pituudet
    Lpi_list = [6000, 8000, 10000]  # mm
    # Lpi_list = [6000]  # mm

    # Kehäjako
    c = 6000  # mm
    # Nurjahduspituudet
    lcr_list = [2, 0.7]
    # Heikomman suunnan nurjahdus
    buckling_z = [True, False]
    # Kiepahdus
    LT_buckling = [True, False]
    # Poikkileikkausluokat
    cross_section_class_list = [3, 2]
    # Profiili
    sym = "dual"  # dual or mono
    prob_type = "discrete"  # continuous or discrete

    # Momentti z-akselin ympäri
    Mz = 0
    # Pystysuuntainen tasainen kuorma
    Qy = 0
    # Vaakasuuntainen tasainen kuorma
    Qx = 1.5 * q_wind * c

    # Aloituspiste
    #  x0 = [800, 50, 500, 50, 500, 50]
    #  x0 = [250, 8, 150, 10]
    #  x0 = [340, 6, 180, 10]
    x0 = [300, 8, 200, 10]

    #  x0 = [400, 20, 200, 20]
    #  x0 = [300, 8, 200, 10, 200, 10]
    #  x0 = [var.ub for var in problem.vars]

    # Laskentatapauksien läpikäyminen ja optimointitehtävien ratkaiseminen
    for lcr in lcr_list:
        for cross_section_class in cross_section_class_list:
            for Lpi in Lpi_list:
                for L in L_list:
                    for LT in LT_buckling:
                        for buck_z in buckling_z:
                            Fy = -(1.15 * (g_truss + g_roof * c) + (
                                    1.5 * q_snow * c)) * L / 2

                            # phi_0 = 1 / 200
                            # m = 2
                            # alpha_m = math.sqrt(0.5 * (1 + 1 / m))
                            # alpha_h = max(2 / 3, min(2 / math.sqrt(Lpi), 1))
                            # phi = phi_0 * alpha_m * alpha_h

                            phi = en1993_1_1.sway_imperfection(Lpi, m=2)

                            Fx = phi * (-Fy)

                            problem = WIColumn(
                                L, Lpi, Fx, Fy, Qx, Qy, Mz, lcr=lcr,
                                top_flange_class=cross_section_class,
                                bottom_flange_class=cross_section_class,
                                web_class=cross_section_class, symmetry=sym,
                                buckling_z=buck_z, LT_buckling=LT,
                                prob_type=prob_type)

                            # problem.all_vars.clear()
                            #
                            # for mem in problem.structure.members.values():
                            #     mem.profile = "IPE 300"
                            # print(problem.eval_cons([]))

                            print("Constraints:", len(problem.cons))
                            print("Nonlinear Constraints:", problem.nnonlincons())
                            
                            # problem([340, 10, 150, 8])

                            # problem = WIColumn(
                            #     L=24000, Lpi=6000, Fx=782.89, Fy=-271200,
                            #     Qx=5.85, Qy=0, Mz=0, lcr=2,
                            #     top_flange_class=3,
                            #     bottom_flange_class=3,
                            #     web_class=3,
                            #     symmetry="dual",
                            #     buckling_z=True,
                            #     LT_buckling=True,
                            #     prob_type='continuous')

                            # print("L={0}, Lpi={1}, Fx={2}, Fy={3}, Qx={4}, "
                            #       "Qy={5}, Mz={6}, lcr={7}, "
                            #       "cross_section_class={8}, sym={9}, "
                            #       "buckling_z={10}, LT={11}, prob_type={12}"
                            #       .format(L, Lpi, Fx, Fy, Qx, Qy, Mz, lcr,
                            #               cross_section_class, sym, buck_z,
                            #               LT, prob_type))

                            #vnew = deepcopy(problem.vars[0])
                            #solver = TrustRegionConstr()
                            #f_best, x_best, nit = solver.solve(
                            #    problem, maxiter=200, x0=x0)

                            #problem.num_iters = nit
                            #problem(x_best, prec=5)

                            # BnB
                            lb_solver = TrustRegionConstr()
                            # lb_solver = slp.SLP()
                            solver = BnB(problem, lb_solver)

                            # return solver, problem

                            # break

                            solver.solve(problem, x0=x0, verb=2)

                            x0 = solver.best_x
                            print("x0=", x0)
                            print("Solver done.")
                            #print(solver.X)
                            print("Best found solution.",solver.best_x)
                            print("Best found objective function value:",solver.best_f)
                                                
                            # problem(solver.best_x)

                            #return solver, problem

                            # TrustRegionConstr
                            # solver = TrustRegionConstr()
                            # f_best, x_best, nit = solver.solve(
                            #     problem,
                            #     maxiter=200,
                            #     x0=x0)
                            # problem.num_iters = nit
                            # problem(x_best, prec=5)

                            # SLP
                            # solver = slp.SLP(move_limits=[0.9, 6])
                            # solver.solve(problem,
                            #              maxiter=500,
                            #              maxtime=40,
                            #              x0=x0)
                            # problem(solver.X, prec=5)

                            # SLSQP
                            # solver = slsqp.SLSQP()
                            # f_best, x_best = solver.solve(problem,
                            #                               maxiter=100,
                            #                               x0=x0)
                            # problem(solver.best_x, prec=5)

                            # MISLP
                            # solver = mislp.MISLP(move_limits=[0.5, 5])
                            # # problem(x0)
                            # solver.solve(problem,
                            #              maxiter=200,
                            #              x0=x0,
                            #              min_diff=1e-2,
                            #              verb=True)
                            # problem(solver.X, prec=5)

                            # 2-vaihetekniikalla
                            # solver1 = SLP(move_limits=[0.9, 4])
                            # solver1 = TrustRegionConstr()
                            # solver2 = mislp.MISLP(move_limits=[0.85, 1.5])
                            # solver = two_phase.TwoPhase(
                            #     solver1, solver2, limits=[3, 3])
                            # fopt, xopt = solver.solve(problem,
                            #                           x0=x0,
                            #                           maxiter=200,
                            #                           # min_diff=1e-6,
                            #                           # verb=True
                            #                           )
                            # problem(xopt)
                            # ResultExporter(problem, solver2).to_csv()
                            # ResultExporter(problem, solver2).csv_to_excel()

                            # # ResultExporter muille kuin 2-vaihetekniikalle
                            # ResultExporter(problem, solver).to_csv()
                            # ResultExporter(problem, solver).csv_to_excel()

                            seconds = time.process_time()
                            m, s = divmod(seconds, 60)
                            h, m = divmod(m, 60)

                            print("")
                            print("Process time:", "%d:%02d:%02d" % (h, m, s))
                            print("----------------------------")
                            print("")

                            # breakpoint()

                            #  wi.cross_section.draw()


# if __name__ == '__main__':
#
#     solver, p = ColumnCalculation()
#
#     p.vars[1].lock(6)



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
from metku.truss2d import *
from metku.frame2d.frame2d import *

import numpy as np

# Simple frame parameters
num_storeys = 1
num_bays = 1
storey_H = 5000
bay_L = 12000

# Simple truss parameters
H0 = storey_H
H1 = 1000
H2 = 1500
H3 = 1000
L1 = bay_L / 2
L2 = bay_L / 2
n = 16

dx = 500

# Create simple frame without beams
frame = Frame2D(simple=[num_storeys, num_bays, storey_H, bay_L],
                supports='fixed',
                beams=False)


# Create truss
truss = Truss2D(simple={'H0': H0,
                        'H1': H1,
                        'H2': H2,
                        'L1': L1,
                        'L2': L2,
                        'n': n,
                        'dx': dx})

frame.add(truss)

for joint in truss.joints.values():
    joint.loc += np.random.randint(-1, 2) * 1e-1
    joint.loc = max(0, joint.loc)
    joint.loc = min(1, joint.loc)


frame.plot(print_text=False)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 15 18:16:56 2018

Euler-Bernoulli beam in 2D with rotational stiffness at nodes

@author: kmela
"""

import numpy as np
from functools import lru_cache

CACHE_BOUND = 2**10
from .ebbeam import EBBeam
try:
    from metku.framefem import LineLoad
except:
    from framefem import LineLoad

# Constants
kRigid = 1e20  # articifial large stiffness for rigid joints
kHinge = 1e-4  # artificial small stiffness for hinges


class EBSemiRigidBeam(EBBeam):
    """ Euler-Bernoulli beam element with possibility to include
        semi-rigid joints. Based on Jalkanen (2004):
        JOUSTAVASTI TUETTU TASOPALKKIELEMENTTI, Journal of Structural Mechanics


        Parameters:
        -----------

        :param n1: element's first node
        :param n2: element's second node
        :param section: element's section
        :param material: element's material
        :param rot_stiff: element's ends rotational stiffness [n1 , n2] [kNm/rad]

        :type n1: FEMNode
        :type n2:  FEMNode
        :type section: Section
        :type material: Material
        :type rot_stiff: list

        Variables:
        ----------
        
        :ivar rot_stiff:
            
        Check EBBeam for more info

    """

    def __init__(self, n1, n2, section, material, rot_stiff=[np.inf, np.inf]):
        """

        """
        EBBeam.__init__(self, n1, n2, section, material)

        self.bending_moment = [0.0, 0.0]
        self.shear_force = [0.0, 0.0]
        for i in range(2):
            if rot_stiff[i] <= kHinge:
                rot_stiff[i] = kHinge
            elif rot_stiff[i] >= kRigid:
                rot_stiff[i] = kRigid

        # print(rot_stiff)
        self.rot_stiff = rot_stiff

    def transformation_matrix(self):
        """ From local to global coordinates """
        c = self.direction_cosines()
        L = np.zeros((6, 6))
        L[0, 0:2] = c[0:2]
        L[1, 0] = -L[0, 1]
        L[1, 1] = L[0, 0]
        L[2, 2] = 1.0
        L[3:6, 3:6] = L[0:3, 0:3]
        return L

    def gamma(self):
        """ Eq. (12) of Jalkanen (2004) """

        gamma = []

        E = self.material.young
        I1 = self.section.I[0]
        L = self.length()
        g0 = E * I1 / L

        for k in self.rot_stiff:
            """ k = 0 means a hinged connection """
            if k == 0.0:
                gamma.append(np.inf)
            elif k == kRigid:
                gamma.append(0.0)
            else:
                gamma.append(g0 / k)
        return gamma

    def delta(self):
        """ Eq. (15) of Jalkanen (2004) """
        gamma = self.gamma()
        delta = (1 + 4 * gamma[0]) * (1 + 4 * gamma[1]) - 4 * gamma[0] * gamma[1]
        return delta

    def matrix_S(self):
        """ Eq. (16) of Jalkanen (2004) """
        S = np.zeros((4, 4))
        L = self.length()
        gamma = self.gamma()

        S[1, 0:4] = [-6 * gamma[0] / L * (1 + 2 * gamma[1]), -4 * gamma[0] * (1 + 3 * gamma[1]),
                     6 * gamma[0] / L * (1 + 2 * gamma[1]), -2 * gamma[0]]
        S[3, 0:4] = [-6 * gamma[1] / L * (1 + 2 * gamma[0]), -2 * gamma[1], 6 * gamma[1] / L * (1 + 2 * gamma[0]),
                     -4 * gamma[1] * (1 + 3 * gamma[0])]

        return S

    def global_dofs(self):
        """ Get numbering of element global degrees of freedom """
        return np.append(self.nodes[0].dofs, self.nodes[1].dofs)
        #return self.nodes[0].dofs + self.nodes[1].dofs

    @lru_cache(CACHE_BOUND)
    def local_stiffness_matrix(self, E, A, I1, Le):
        """ Stiffness matrix in local coordinates """

        rodc = E * A / Le
        EI1 = E * I1
        bend1 = 12 * EI1 / Le ** 3
        bend2 = 6 * EI1 / Le ** 2
        bend3 = 4 * EI1 / Le
        bend4 = 2 * EI1 / Le

        # k00 = np.zeros((4, 4))
        k0 = np.zeros((6, 6))

        k00 = np.array([[bend1, bend2, -bend1, bend2],
                        [bend2, bend3, -bend2, bend4],
                        [-bend1, -bend2, bend1, -bend2],
                        [bend2, bend4, -bend2, bend3]])


        delta = self.delta()
        S = self.matrix_S()

        k1 = 1 / delta * k00.dot(S) + \
             1 / delta * S.transpose().dot(k00) + \
             1 / delta ** 2 * S.transpose().dot(k00).dot(S)

        # s2 = S[1,:]
        # s4 = S[3,:]
        k2 = np.zeros((4, 4))
        for i in range(0, 2):
            if self.rot_stiff[i] < kRigid:
                s = S[2 * i + 1, :]
                # print(s)
                k2 += self.rot_stiff[i] / delta ** 2 * np.outer(s, s)

        # k2 = self.rot_stiff[0]/delta**2*np.outer(s2,s2) + self.rot_stiff[1]/delta**2*np.outer(s4,s4)


        # print(k00)
        # print(k1)
        # print(k2)

        k0[np.ix_([1, 2, 4, 5], [1, 2, 4, 5])] = k00 + k1 + k2
        k0[[0, 3], [0, 3]] = rodc
        k0[[0, 3], [3, 0]] = -rodc

        return k0

    def stiffness_matrix(self):
        """ Compute the stiffness matrix """

        E = self.material.young
        A = self.section.A
        I1 = self.section.I[0]
        Le = self.length()
        k0 = self.local_stiffness_matrix(E, A, I1, Le)

        # k0 = CheckReleases[fem.elem[N],k0];

        # local-global matrix
        L = self.transformation_matrix()

        # globaali elementin matriisi
        ke = L.transpose().dot(k0.dot(L))

        return ke

    def local_geometric_stiffness_matrix(self,lcase=0):
        """ Geometric stiffness matrix in local coordinates
            From: Cook et. al 1989, Section 14.2
        """
        #P = self.axial_force[1]
        P = self.fint[lcase]['fx'][1]
        Le = self.length()

        g0 = P / 30 / Le
        keg0 = g0 * np.array([[36, 3 * Le, -36, 3 * Le],
                              [3 * Le, 4 * Le ** 2, -3 * Le, -Le ** 2],
                              [-36, -3 * Le, 36, -3 * Le],
                              [3 * Le, -Le ** 2, -3 * Le, 4 * Le ** 2]])
        kg0 = np.zeros([6, 6])
        q = [1, 2, 4, 5]
        kg0[np.ix_(q, q)] = keg0

        return kg0

    def geometric_stiffness_matrix(self,lcase=0):
        """ Geometric stiffness matrix in global coordinates
            From: Cook et. al (1989), Section 14.2
        """

        # local-global transformation matrix
        L = self.transformation_matrix()

        kG0 = self.local_geometric_stiffness_matrix(lcase)

        kG = L.transpose().dot(kG0.dot(L))

        return kG

    def equivalent_nodal_loads(self, load):
        """ Equivalent nodal loads for a load

            'load' is a Load type object. By default, it is assumed that
            'load' is LineLoad

            Returns:
            ---------
            :return: Equivalent nodal loads for load in vector q
            :rtype: np.array
        """
        floc = np.zeros(6)

        T = self.transformation_matrix()
        L = self.length()

        if isinstance(load, LineLoad):
            q1 = load.qval[0]
            q2 = load.qval[1]
            # print(q1,q2)
            if load.coords == 'local':
                # print('local coordinates')
                """ Load is given in local coordinates """
                if load.dir == "x":
                    """ Load is in the axial direction """
                elif load.dir == 'y':
                    """ Load is perpendicular to the member """
                    floc[1] = 7 / 20 * q1 * L + 3 / 20 * q2 * L
                    floc[4] = 3 / 20 * q1 * L + 7 / 20 * q2 * L
                    floc[2] = L ** 2 * (1 / 20 * q1 + 1 / 30 * q2)
                    floc[5] = -L ** 2 * (1 / 30 * q1 + 1 / 20 * q2)
                # print(floc)
            else:
                """ Load is given in global coordinates: it has to be transformed
                    first into local coordinates.
                """
                if load.dir == "x":
                    # q = np.array([q1, 0, 0])
                    q = np.array([1, 0, 0])
                else:
                    # q = np.array([0, q1, 0])
                    q = np.array([0, 1, 0])
                # Load vector transformed into element local coordinate system
                qloc = T[:3, :3].dot(q)

                """ Loads perpendicular to the axis of the element """
                qy1 = q1 * qloc[1]
                qy2 = q2 * qloc[1]
                floc[1] = 7 / 20 * qy1 * L + 3 / 20 * qy2 * L
                floc[4] = 3 / 20 * qy1 * L + 7 / 20 * qy2 * L
                floc[2] = L ** 2 * (1 / 20 * qy1 + 1 / 30 * qy2)
                floc[5] = -L ** 2 * (1 / 30 * qy1 + 1 / 20 * qy2)

                """ Loads parallel to the axis of the element """
                qx1 = q1 * qloc[0]
                qx2 = q2 * qloc[0]
                floc[0] = 0.5 * L * (0.5 * (qx1 + qx2) - 1 / 6 * (qx2 - qx1))
                floc[3] = 0.5 * L * (0.5 * (qx1 + qx2) + 1 / 6 * (qx2 - qx1))

                # Construct nodal load in local coordinates
                # qloc[0,1,2] = axial, shear, moment of node 1
                # qloc[3,4,5] = axial, shear, moment ofnode 2
                # Axial force
                # floc[[0, 3]] = 0.5 * qloc[0] * L
                # Shear force
                # floc[[1, 4]] = 0.5 * qloc[1] * L
                # Moment
                # floc[2] = qloc[1] * L ** 2 / 12.0
                # floc[5] = -floc[2]

                # print(floc)

        if len(self.releases) > 0:
            """ There are releases in the element, so the local load vector
                is modified.
            """
            rel = self.releases
            # nrel is the list of non-released forces
            nrel = np.setdiff1d(np.arange(6), rel)

            # print(np.linalg.inv(self.Krel['K22']).dot(floc[rel]))
            floc[nrel] -= self.Krel['K12'].dot(np.linalg.inv(self.Krel['K22']).dot(floc[rel]))
            floc[rel] = 0

            # Store local loads
        self.floc[load.sid] = floc

        # print(self.floc)

        fglob = T.transpose().dot(floc)

        return fglob


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue 30 Mar 2021

Solving mixed-integer linear programming problems.

@author: kmela
"""

# TODO! Add a MILP solver from ortools

from scipy.optimize import linprog
import numpy as np
import time

# import gurobipy as grb

from metku.optimization.solvers.optsolver import OptSolver
import metku.optimization.structopt as sopt

#from ortools.linear_solver import pywraplp

class MILP(OptSolver):

    def __init__(self,algorithm="gurobi"):
        super().__init__()
        self.algo = algorithm
        

    def solve(self,problem,verb=False):
        """
        
        """
    
        import gurobipy as grb
                
        lp = grb.Model("milp")
        
                
        x = []
        for var in problem.vars:
            """
                create variable, including the objective function
                coefficient and name
            """
            if isinstance(var,sopt.BinaryVariable):
                x.append(lp.addVar(var.lb,var.ub,vtype=grb.GRB.BINARY,name=var.name))
            else:
                x.append(lp.addVar(var.lb,var.ub,vtype=grb.GRB.CONTINUOUS,name=var.name))
                
        #nvars = len(x)
                
        if problem.obj.obj_type == "MIN":
            obj_sense = grb.GRB.MINIMIZE
        else:
            obj_sense = grb.GRB.MAXIMIZE
        
        lp.setObjective(grb.LinExpr(problem.obj.c,x),sense=obj_sense)
        
        
        for con in problem.cons:
            if isinstance(con,sopt.LinearConstraint):
                if con.type == '<':
                    con_sense = grb.GRB.LESS_EQUAL
                            #lp.addLConstr(LinExpr(con.a, x), grb.GRB.LESS_EQUAL, con.b)
                            #lp.addConstr(grb.quicksum([con.a[j]*x[j] for j in range(nvars)]) <= con.b[i])
                elif con.type == '>':
                    con_sense = grb.GRB.GREATER_EQUAL
                            #lp.addLConstr(LinExpr(con.a, x), grb.GRB.LESS_EQUAL, con.b)
                            #lp.addConstr(grb.quicksum([con.a[j]*x[j] for j in range(nvars)]) >= con.b[i])
                else:
                    con_sense = grb.GRB.EQUAL
                            #lp.addConstr(grb.quicksum([con.a[j]*x[j] for j in range(nvars)]) = con.b[i])
                
                lp.addLConstr(grb.LinExpr(con.a, x), con_sense, con.b)
                

        lp.update()                
       

        #for var in lp.getVars():
        #    print(var.lb,var.ub)
        
        
        if not verb:
            lp.setParam("LogToConsole",0)
                #qp.update()
        lp.optimize()
                
        if lp.Status == grb.GRB.OPTIMAL:
            self.feasible = True
        elif lp.Status == grb.GRB.INFEASIBLE:
            self.feasible = False
                
        X = []
        for v in lp.getVars():
            #print('%s %g' % (v.varName, v.x))                    
            X.append(v.x)
        
        # bounds = [x*self.move_limits for x in self.X]
        #
        # res = linprog(df, A, B, bounds=bounds)

        self.X = X
        #return X


# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
import os
from frame2d.frame2d import SteelBeam

def to_robot_3D(filename,
                end_frame1,
                long_frame1,
                main_frame,
                num_frames,
                s,
                L,
                end_frame2=None,
                long_frame2=None,
                stiff_truss=None):
        """  Creates an Aurodesk Robot Structural Analysis .str -file
        
            Parameters
            ----------
            :param filename: Name of the created file
            :param num_frames: Number of frames 
            :param s: Spacing between created frames
            :param brace_profile: Profile for vertical braces
                
            :type filename: string
            :type num_frames: int
            :type s: float
            :type brace_profile: string
        """
        if not end_frame2:
            end_frame2 = end_frame1
        if not long_frame2:
            long_frame2 = long_frame1
        num_frames -= 1
        L1=L/2
        L2=L/2
        nodes = []
        elements = []
        profiles = []
        material = []
        releases = {}
        pointloads = []
        #vert_braces = []
        # 2*end_frames + 2*long_frames + 2 = 5
        for i in range(num_frames+7):
            if i == 0:
                frame = end_frame1                
            elif i == num_frames+2:
                frame = end_frame2
            elif i == num_frames+3:
                frame = long_frame1
            elif i == num_frames+4:
                frame = long_frame2
            elif i == num_frames+5:
                if stiff_truss:
                    frame = stiff_truss
                    # Omit eccentricity elements
                    frame.truss = False 
                else:
                    continue
            elif i == num_frames+6:
                if stiff_truss:
                    frame = stiff_truss
                    # Omit eccentricity elements
                    frame.truss = False 
                else:
                    continue
            else:
                frame = main_frame
            
            # Loop through members
            for member in frame.members.values():
                n1, n2 = member.coordinates
                if i == num_frames+3:
                    if isinstance(member, SteelBeam):
                        n1 = [0, n1[0], n1[1]]
                        n2 = [0, n2[0], n2[1]]
                    else:
                        continue
                elif i == num_frames+4:
                    if isinstance(member, SteelBeam):
                        n1 = [L, n1[0], n1[1]]
                        n2 = [L, n2[0], n2[1]]
                    else:
                        continue
               # Wind truss 
                elif i == num_frames+5:                   
                    if member.mtype=='web':
                        idx1 = int(abs(n1[0]-1e-3) // (main_frame.truss[0].L1))
                        idx2 = int(abs(n2[0]-1e-3)//  (main_frame.truss[0].L2))
                        idxT1 = int(idx1 // 2)
                        idxT2 = int(idx2 // 2)
                        idx1 -= 2*idxT1
                        idx2 -= 2*idxT2
                        if main_frame.truss:   
                            chord1 = main_frame.truss[idxT1].top_chords[idx1]
                            chord2 = main_frame.truss[idxT2].top_chords[idx2]
                        else:
                            beams = [b for b in main_frame.members.values() if b.mtype == "beam"]
                            chord1 = beams[idx1]
                            chord2 = beams[idx2]
                        n1 = [n1[0], 0, chord1.shape(n1[0])]
                        n2 = [n2[0], s, chord2.shape(n2[0])]                
                    else:
                        continue
                # Wind truss
                elif i == num_frames+6:
                    if member.mtype=='web':
                        idx1 = int(abs(n1[0]-1e-3) //  (main_frame.truss[0].L1))
                        idx2 = int(abs(n2[0]-1e-3)//  (main_frame.truss[0].L2))
                        idxT1 = int(idx1 // 2)
                        idxT2 = int(idx2 // 2)
                        idx1 -= 2*idxT1
                        idx2 -= 2*idxT2
                        if main_frame.truss:   
                            chord1 = main_frame.truss[idxT1].top_chords[idx1]
                            chord2 = main_frame.truss[idxT2].top_chords[idx2]
                        else:
                            beams = [b for b in main_frame.members.values() if b.mtype == "beam"]
                            chord1 = beams[idx1]
                            chord2 = beams[idx2]
                        n1 = [n1[0], s*(num_frames+2), chord1.shape(n1[0])]
                        n2 = [n2[0], s*(num_frames+1), chord2.shape(n2[0])]                                    
                    else:
                        continue
                else:
                    n1 = [n1[0], i*s, n1[1]]
                    n2 = [n2[0], i*s, n2[1]]
                    
                if n1 not in nodes:
                    nodes.append(n1)
                if n2 not in nodes:
                    nodes.append(n2)
                elem = [nodes.index(n1) + 1, nodes.index(n2) + 1]
                elements.append(elem)
                splitted_val = member.profile.split(" ")
                profile_type = splitted_val[0]
                """
                # Add columns end node to vertical braces list
                if num_frames != 1 and member.mtype == 'top_chord':
                    if nodes.index(n1) +1 not in vert_braces:
                        vert_braces.append(nodes.index(n1) +1)
                    
                    if nodes.index(n2) +1 not in vert_braces:
                        vert_braces.append(nodes.index(n2) +1)
                """
                
                if profile_type == 'RHS':
                    profile = 'RRHS ' + splitted_val[1]
                elif profile_type == 'HE':
                    profile = splitted_val[0] + splitted_val[2] + splitted_val[1]
                elif profile_type == "SHS":
                    dims = splitted_val[1].split("X")
                    profile = 'RRHS ' + str(dims[0]) + 'X' + str(dims[0]) + 'X' + str(dims[1])
                else:
                    profile = member.profile
                # Webs
                if member.mtype == "web" or member.mtype == "beam" or\
                    member.mtype=="top_chord" or member.mtype == "bottom_chord":
                    Sj1 = min(1e10, member.Sj1)
                    Sj2 = min(1e10, member.Sj2)
                    releases[elements.index(elem) +1] = f'ORIgin RY Hy={Sj1}  END RY Hy={Sj2}'
    
                profiles.append(profile)
                material.append(member.material)
                
                # Eccentricity elements
                if len(member.ecc_coordinates):
                    for coords in member.ecc_coordinates:
                        n1, n2 = coords
                        if num_frames != 1:
                            n1 = [n1[0], i*s, n1[1]]
                            n2 = [n2[0], i*s, n2[1]]
                        if n1 not in nodes:
                            nodes.append(n1)
                        if n2 not in nodes:
                            nodes.append(n2)
                        elem = [nodes.index(n1) + 1, nodes.index(n2) + 1]
                        elements.append(elem)
                        profiles.append("HEA 1000")
                        material.append("S355")
                        #if n1 < n2:
                        #    releases[elements.index(elem) +1] = 'END RY'
                        #else:
                        #    releases[elements.index(elem) +1] = 'ORIgin RY'
                        
                        
    
            
            if frame.truss:
                for truss in frame.truss:
                    for joint in truss.joints.values():
                        # Y joint
                        if len(joint.nodal_coordinates) == 2:
                            n1, n2 = sorted(joint.nodal_coordinates)
                            if num_frames != 1:
                                n1 = [n1[0], i*s, n1[1]]
                                n2 = [n2[0], i*s, n2[1]]
                            if n1 not in nodes:
                                nodes.append(n1)
                            if n2 not in nodes:
                                nodes.append(n2)
                            # Eccentricity element
                            elem = [nodes.index(n1) + 1, nodes.index(n2) + 1]
                            elements.append(elem)
                            profiles.append("HEA 1000")
                            material.append("S355")
                        # K joint
                        elif len(joint.nodal_coordinates) == 5:
                            coords = sorted(joint.nodal_coordinates)
                            n1, n2, n3, n4, n5 = coords
                            if num_frames != 1:
                                n1 = [n1[0], i*s, n1[1]]
                                n2 = [n2[0], i*s, n2[1]]
                                n3 = [n3[0], i*s, n3[1]]
                                n4 = [n4[0], i*s, n4[1]]
                                n5 = [n5[0], i*s, n5[1]]
                            if n1 not in nodes:
                                nodes.append(n1)
                            if n2 not in nodes:
                                nodes.append(n2)
                            if n4 not in nodes:
                                nodes.append(n4)
                            if n5 not in nodes:
                                nodes.append(n5)
                            # Eccentricity elements
                            elem1 = [nodes.index(n1) + 1, nodes.index(n2) + 1]
                            elem2 = [nodes.index(n4) + 1, nodes.index(n5) + 1]
                            elements.append(elem1)
                            elements.append(elem2)
                            profiles.append("HEA 1000")
                            material.append("S355")
                            profiles.append("HEA 1000")
                            material.append("S355")
                            #releases[elements.index(elem1)+1] = "END RY"
                            #releases[elements.index(elem2)+1] = "END RY"
                    else:
                        pass
                    
                    
            for pointload in frame.point_loads.values():
                try:
                    n1 = pointload.coordinate
                    n1 = [n1[0], i*s, n1[1]]
                    idx = nodes.index(n1) + 1
                except ValueError:
                    n1 = pointload.coordinate
                    n1 = [n1[0], i*s, n1[1]]
                    nodes.append(n1)
                    idx = nodes.index(n1) +1
                FX, FZ, MY = pointload.v
                if FX:
                    pointloads.append(f'    {idx} FX={FX}')
                elif FX and FZ:
                    pointloads.append(f'    {idx} FX={FX}  FZ={FZ}')
                elif FX and MY:
                    pointloads.append(f'    {idx} FX={FX}  MY={MY}')
                elif FZ:
                    pointloads.append(f'    {idx} FZ={FZ}')
                elif FZ and MY:
                    pointloads.append(f'    {idx} FZ={FZ}  MY={MY}')
                elif MY:
                    pointloads.append(f'    {idx} MY={MY}')
                elif FX and FZ and MY:
                    pointloads.append(f'    {idx} FX={FX}  FZ={FZ}  MY={MY}')


        with  open(filename + '.str', 'w') as f:
            f.write("ROBOT97 \n")
            if num_frames != 1:
                f.write("FRAme SPAce \n")
            f.write("NUMbering DIScontinuous \n")
            f.write(f'NODes {len(nodes)}  ELEments {len(elements)} \n')
            f.write("UNIts \n")
            f.write("LENgth=mm	Force=N \n")
            f.write("NODes \n")
            for i, node in enumerate(nodes):
                if num_frames != 1:
                    f.write(f'{i+1}   {node[0]}    {node[1]}    {node[2]} \n')
                else:
                    f.write(f'{i+1}   {node[0]}    {node[1]} \n')

            f.write('ELEments \n')
            for i, element in enumerate(elements):
                f.write(f' {i+1}  {element[0]}    {element[1]} \n')

            f.write('PROperties \n')
            for i in range(len(elements)):
                f.write(f' "{material[i]}" \n')
                f.write(f' {i+1}  ')
                f.write(f' {profiles[i]}    \n')

            f.write("SUPports \n")
            # End frames
            if end_frame1.is_calculated:
                for sup in end_frame1.supports.values():
                    n1 = sup.coordinate
                    n1 = [n1[0], 0, n1[1]]
                    idx = nodes.index(n1) + 1
                    if sup.dofs[0] == [-1]:
                        f.write(f'{idx}  UX  \n')
                    elif sup.dofs[0:1] == [-1, -1]:
                        f.write(f'{idx}  UX UZ \n')
                    elif sup.dofs == [-1, -1, -1]:
                        f.write(f'{idx}  \n')
                    elif sup.dofs[1] == [-1]:
                        f.write(f'{idx}  UZ  \n')
                    elif sup.dofs[1:2] == [-1, -1]:
                        f.write(f'{idx}  UZ RY  \n')
                    else:
                        f.write(f'{idx}  RY \n')
            else:
                for sup in end_frame1.supports.values():
                        n1 = sup.coordinate
                        n1 = [n1[0], 0, n1[1]]
                        idx = nodes.index(n1) + 1
                        if sup.dofs == [1, 0, 0]:
                            f.write(f'{idx}  UX  \n')
                        elif sup.dofs == [1, 1, 0]:
                            f.write(f'{idx}  UX UZ \n')
                        elif sup.dofs == [1, 1, 1]:
                            f.write(f'{idx}  \n')
                        elif sup.dofs == [0, 1, 0]:
                            f.write(f'{idx}  UZ  \n')
                        elif sup.dofs == [0, 1, 1]:
                            f.write(f'{idx}  UZ RY  \n')
                        else:
                            f.write(f'{idx}  RY \n')
            if end_frame2.is_calculated:
                for sup in end_frame2.supports.values():
                    n1 = sup.coordinate
                    n1 = [n1[0], (num_frames+2)*s, n1[1]]
                    idx = nodes.index(n1) + 1
                    if sup.dofs[0] == [-1]:
                        f.write(f'{idx}  UX  \n')
                    elif sup.dofs[0:1] == [-1, -1]:
                        f.write(f'{idx}  UX UZ \n')
                    elif sup.dofs == [-1, -1, -1]:
                        f.write(f'{idx}  \n')
                    elif sup.dofs[1] == [-1]:
                        f.write(f'{idx}  UZ  \n')
                    elif sup.dofs[1:2] == [-1, -1]:
                        f.write(f'{idx}  UZ RY  \n')
                    else:
                        f.write(f'{idx}  RY \n')
            else:
                for sup in end_frame2.supports.values():
                        n1 = sup.coordinate
                        n1 = [n1[0], (num_frames+2)*s, n1[1]]
                        idx = nodes.index(n1) + 1
                        if sup.dofs == [1, 0, 0]:
                            f.write(f'{idx}  UX  \n')
                        elif sup.dofs == [1, 1, 0]:
                            f.write(f'{idx}  UX UZ \n')
                        elif sup.dofs == [1, 1, 1]:
                            f.write(f'{idx}  \n')
                        elif sup.dofs == [0, 1, 0]:
                            f.write(f'{idx}  UZ  \n')
                        elif sup.dofs == [0, 1, 1]:
                            f.write(f'{idx}  UZ RY  \n')
                        else:
                            f.write(f'{idx}  RY \n')
                            
            # Longitudal frames
            if long_frame1.is_calculated:
                for sup in long_frame1.supports.values():
                    n1 = sup.coordinate
                    n1 = [n1[1], n1[0], 0]
                    idx = nodes.index(n1) + 1
                    if sup.dofs[0] == [-1]:
                        f.write(f'{idx}  UX  \n')
                    elif sup.dofs[0:1] == [-1, -1]:
                        f.write(f'{idx}  UX UZ \n')
                    elif sup.dofs == [-1, -1, -1]:
                        f.write(f'{idx}  \n')
                    elif sup.dofs[1] == [-1]:
                        f.write(f'{idx}  UZ  \n')
                    elif sup.dofs[1:2] == [-1, -1]:
                        f.write(f'{idx}  UZ RY  \n')
                    else:
                        f.write(f'{idx}  RY \n')
            else:
                for sup in long_frame1.supports.values():
                        n1 = sup.coordinate
                        n1 = [n1[1], n1[0], 0]
                        idx = nodes.index(n1) + 1
                        if sup.dofs == [1, 0, 0]:
                            f.write(f'{idx}  UX  \n')
                        elif sup.dofs == [1, 1, 0]:
                            f.write(f'{idx}  UX UZ \n')
                        elif sup.dofs == [1, 1, 1]:
                            f.write(f'{idx}  \n')
                        elif sup.dofs == [0, 1, 0]:
                            f.write(f'{idx}  UZ  \n')
                        elif sup.dofs == [0, 1, 1]:
                            f.write(f'{idx}  UZ RY  \n')
                        else:
                            f.write(f'{idx}  RY \n')
                            
            if long_frame2.is_calculated:
                for sup in long_frame2.supports.values():
                    n1 = sup.coordinate
                    n1 = [L, n1[0], 0]
                    idx = nodes.index(n1) + 1
                    if sup.dofs[0] == [-1]:
                        f.write(f'{idx}  UX  \n')
                    elif sup.dofs[0:1] == [-1, -1]:
                        f.write(f'{idx}  UX UZ \n')
                    elif sup.dofs == [-1, -1, -1]:
                        f.write(f'{idx}  \n')
                    elif sup.dofs[1] == [-1]:
                        f.write(f'{idx}  UZ  \n')
                    elif sup.dofs[1:2] == [-1, -1]:
                        f.write(f'{idx}  UZ RY  \n')
                    else:
                        f.write(f'{idx}  RY \n')
            else:
                for sup in long_frame2.supports.values():
                        n1 = sup.coordinate
                        n1 = [L, n1[0], 0]
                        idx = nodes.index(n1) + 1
                        if sup.dofs == [1, 0, 0]:
                            f.write(f'{idx}  UX  \n')
                        elif sup.dofs == [1, 1, 0]:
                            f.write(f'{idx}  UX UZ \n')
                        elif sup.dofs == [1, 1, 1]:
                            f.write(f'{idx}  \n')
                        elif sup.dofs == [0, 1, 0]:
                            f.write(f'{idx}  UZ  \n')
                        elif sup.dofs == [0, 1, 1]:
                            f.write(f'{idx}  UZ RY  \n')
                        else:
                            f.write(f'{idx}  RY \n')
            
            f.write("RELeases \n")
            for elem in releases.keys():
                f.write(f'ELEments{elem} {releases[elem]} \n')
            
            f.write("LOAds \n")
            f.write("CASe # 1 LC1 \n")
            if len(main_frame.line_loads):
                f.write("ELEments \n")
                for i in range(num_frames):
                    for lineload in main_frame.line_loads.values():
                        n1, n2 = lineload.member.coordinates
                        if num_frames != 1:
                            n1 = [n1[0], i*s, n1[1]]
                            n2 = [n2[0], i*s, n2[1]]
                        n1_idx = nodes.index(n1) + 1
                        n2_idx = nodes.index(n2) + 1
                        idx = elements.index([n1_idx, n2_idx]) + 1
                        if lineload.direction == 'y':
                            dir = 'PZ'
                        else:
                            dir = 'PX'
                        q0, q1 = lineload.values
                        if q0 != q1:
                            f.write(f' {idx} X=0.0 {dir}={q0} TILl  ')
                            f.write(f'X=1.000  {dir}={q1}      RElative \n')
        
                        else:
                            f.write(f' {idx} {dir}={q0} \n')
            if len(pointloads):
                f.write("NODes \n")
                for val in pointloads:
                    f.write(val + "\n")
                    

            f.write("END")
        
        print(f'{filename}.str created to: \n{os.getcwd()}')
        
        
        
        
        
        



from enum import Enum, auto, StrEnum

class MemberTypeEnum(StrEnum):
    COLUMN = auto()
    BEAM = auto()
    BRACE = auto()
class SupportMethodEnum(StrEnum):
    REM = auto()
    ZERO = auto()

class LoadCaseEnum(StrEnum):

    PERMANENT = auto()
    IMPOSED_A = auto()
    IMPOSED_B = auto()
    IMPOSED_C = auto()
    IMPOSED_D = auto()
    IMPOSED_E = auto()
    SNOW = auto()
    WIND = auto()
class LoadCombinationEnum(StrEnum):
    ULS = auto()
    SLS_CH = auto()
    SLS_FR = auto()
    SLS_QP = auto()
    FIRE = auto()
    ACC = auto()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2022. Metku team.
#  All rights reserved.

# @Filename : cross_section
# @Date : 2022
# @Project: metku
# @AUTHOR : Jaakko Huusko

from dataclasses import dataclass


@dataclass
class CrossSection:

    material: object = None
    profile: object = None

    # Forces
    N: float = 0
    Vy: float = 0
    Vz: float = 0
    MT: float = 0
    My: float = 0
    Mz: float = 0



from .optsolver import OptSolver
from .slsqp import SLSQP, COBYLA
from .two_phase import TwoPhase
from .ga import GA
#from .optsolver import DiscreteVNS
from .slp import SLP
from .mislp import MISLP
from .vns import VNS, DiscreteVNS
from .trust_region import TrustRegionConstr
from .pso import PSO
from .bnb import BnB
from .mma import MMA


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
class BeamSection(Section):
    """ Class for beam element cross sections

        Parameters:
        -----------
        :param A: cross-sectional area [mm^2]
        :param Iy: second moment of area [mm^4]

        :type A: float
        :type Iy: float

        Variables:
        ----------
        :ivar Iy: second moment of area [mm^4]

        :vartype Iy: float
    """

    def __init__(self, A, Iy):
        Section.__init__(self, A)
        self.Iy = Iy
        """ Second moment of area [mm^4]"""



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Mon Jan 29 20:13:49 2018

@author: kmela

Class for steel beam

"""

import sys
import math
import numpy as np

import eurocode3
from cross_section import CrossSection

class SteelBeam:
    
    def __init__(self,profile,spans=[6.0e3],supports=["Hinge","X"]):
        """ Constructor
            
            profile -- member profile (of cross_section class)
            spans -- list of spans
            supports -- list of support conditions. Possible values are:
                    "Free" .. no support
                    "Hinge" .. x- and y- displacements fixed
                    "X" .. x-direction fixed
                    "Y" .. y-direction fixed
                    "Rigid" .. all degrees of freedom fixed
            
        """
        
        self.profile = profile
        self.spans = spans
        self.supports = supports
        
        self.forces = []
        self.moments = []
        self.loads = []
        
    
    def nspan(self):
        """ Number of spans """
        return len(self.spans)
    
    def length(self):
        """ Beam length """
        return sum(self.spans)
        
    def add_force(x,val):
        """ Adds a point force to the beam
            input: x .. x-coordinate
                val .. magnitude. The sign gives the direction
        """
        
        
    def mesh(self,nel=10):
        """ Generate FEM model 
            input: nel .. number of elements for each span
        """
        
        """ Generate nodes """
        ycoord = 0.0;
        x = 0.0;
        
        fen = FrameFEM()
        
        fem.add_node(x,ycoord)
        
        for span in self.spans:
            dx = span/nel
            for i in range(nel):
                x += dx
                fem.add_node(x,ycoord)

if __name__ == "__main__":


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2023. Metku team.
#  All rights reserved.

from dataclasses import dataclass
import numpy as np


"""
@dataclass
class SIBaseUnit:
    name: str = None
    symbol: str = None
    value: float = 1
    factor: float = 1
    s: float = 0
    m: float = 0
    kg: float = 0
    A: float = 0
    K: float = 0
    mol: float = 0
    cd: float = 0

    @property
    def units(self):
        return self.s, self.m, self.kg, self.A, self.K, self.mol, self.cd

    def as_array(self):
        return np.array(self.units)

    def __add__(self, other):
        if isinstance(other, SIBaseUnit):
            return self.value + other.value

    def __mul__(self, other):
        if isinstance(other, SIBaseUnit):
            self.value *= other.value
            self.as_array()
        return self.value * self.as_array()

    def __rmul__(self, other):
        self.value *= other
        return self.value * self.as_array()
"""

@dataclass
class Unit:
    name: str
    quantity: str
    symbol: str
    base: float

    def __mul__(self, other):
        return other * self.base

    def __rmul__(self, other):
        return other * self.base

    def __str__(self):
        return self.symbol

    def __call__(self, unit):
        return f"{unit / self.base :.2f} {self.symbol}"

"""
@dataclass
class Length:
    mm: float = SIBaseUnit(name="millimeter", symbol="mm", m=1e-3)
    m: float = SIBaseUnit(name="millimeter", symbol="mm", m=1)
"""

@dataclass
class Force:
    N: float = Unit("Newton", "force", "N", 1)
    kN: float = Unit("kiloNewton", "force", "kN", 1e3)
    MN: float = Unit("megaNewton", "force", "MN", 1e6)

"""
@dataclass
class Mass:
    g: float = Unit("gram", "mass", "g", 1e-3)
    kg: float = Unit("kilogram", "mass", "kg", 1)


@dataclass
class Parameters:
    length: Length = Length.mm
    force: Force = Force.N
    mass: Mass = Mass.kg
"""

Settings = Parameters()

if __name__ == '__main__':
    mass = 100 * Length.m
    mass2 = 1000 * Length.mm
    print(mass)



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
from metku.optimization.structopt import OptimizationProblem, Variable, \
    NonLinearConstraint, LinearConstraint, DiscreteVariable
from metku.frame2d.frame2d import FrameMember, Frame2D, SteelBeam, PointLoad, XYHingedSupport
import numpy as np

class ThreeBarTruss(OptimizationProblem):

    # Problem parameters
    P = 20
    L = 100
    sigma_U = 20
    sigma_L = -15

    def __init__(self):
        super().__init__(name='ThreeBarTruss')

        # Problem type needs to be addressed
        # otherwise some solvers might not work
        self.prob_type = 'continuous'
        self.create_variables()
        self.create_constraints()
        self.create_objective()

    def create_variables(self):
        """
        Creates variables
        """
        self.vars = []

        var1 = Variable(name="X1",
                        lb=0.01,
                        ub=5)

        var2 = Variable(name="X2",
                        lb=0.01,
                        ub=5)

        self.vars = [var1, var2]


    def create_objective(self):
        """
        Creates objective function and assigns it to problem
        """

        def objective(X):
            """
            Weight of the truss
            :param X: design variables
            :return: weight
            """
            x1, x2 = X
            L1 = np.sqrt(2) * self.L
            L2 = self.L

            return 2 * L1 * x1 + L2 * x2

        # Assign function to problem
        self.obj = objective

    def create_constraints(self):
        """
        Creates constraints and assigns them to problem
        """
        L1 = np.sqrt(2) * self.L
        L2 = self.L

        self.cons = []

        # Constraint 1
        def g1(x):
            x1, x2 = x
            return self.P * (x2 + np.sqrt(2) * x1) / (
                        2 * x1 * x2 + np.sqrt(2) * x1 ** 2) - self.sigma_U

        # Constraint 2
        def g2(x):
            x1, x2 = x
            return self.P * (np.sqrt(2) * x1) / (
                        2 * x1 * x2 + np.sqrt(2) * x1 ** 2) - self.sigma_U

        # Constraint 3
        def g3(x):
            x1, x2 = x
            return self.P * (x2) / (2 * x1 * x2 + np.sqrt(2) * x1 ** 2) + self.sigma_L

        # Create NonLinearConstraint objects
        con1 = NonLinearConstraint(con_fun=g1, name="Con1", con_type="<")
        con2 = NonLinearConstraint(con_fun=g2, name="Con2", con_type="<")
        con3 = NonLinearConstraint(con_fun=g3, name="Con3", con_type="<")

        # Assign created constraints as a list to the problem
        self.cons = [con1, con2, con3]

if __name__ == '__main__':
    # Import all solvers
    from metku.optimization.solvers import *
    # Create optimization problem
    problem = ThreeBarTruss()
    # Choose solver
    solver = SLP(move_limits=[0.5, 5])
    # Create random initial starting point
    x0 = [np.random.randint(var.lb, var.ub) for var in problem.vars]
    # Solve problem
    # NOTE! if you use SLSQP the order is xopt, fopt
    fopt, xopt = solver.solve(problem, x0=x0, maxiter=500)
    # Print results by calling the problem with the optimal parameters
    problem(xopt)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  This source code is licensed under the MIT license. See LICENSE in the repository root directory.
#  Copyright (c) 2022. Metku team.
#  All rights reserved.

# @Filename : __init__.py
# @Date : 2022
# @Project: metku
# @AUTHOR : Jaakko Huusko
